# Tauon Music Box - Theme reading module

# Copyright © 2015-2022, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

import io
import os
import json
from PIL import Image, ImageDraw, ImageFilter
from sdl2 import *
from sdl2.sdlimage import *
from t_modules.t_extra import *
import base64

def get_colour_from_line(cline):

    colour = ["", "", "", ""]
    mode = 0

    is_hex = False
    if "," not in cline[:8]:
        if cline[:6].isalnum():
            is_hex = True
        if cline[0] == "#" and cline[1:7].isalnum() and cline[7].isspace() and cline[8].isspace():
            is_hex = True
        if cline[0] == "#" and cline[1:9].isalnum() and cline[9].isspace() and cline[10].isspace():
            is_hex = True

    if is_hex:
        # hex mode
        if cline.startswith("0x"):
            cline = cline[2:]
        if cline.startswith("#"):
            cline = cline[1:]
        ll = cline
        a = 255
        r = int(ll[0] + ll[1], 16)
        g = int(ll[2] + ll[3], 16)
        b = int(ll[4] + ll[5], 16)
        if ll[6].isalnum() and ll[7].isalnum():
            a = int(ll[6] + ll[7], 16)
        return [r, g, b, a]
    else:
        # rgb mode
        for i in cline:
            if i.isdigit():
                colour[mode] += i
            elif i == ',':
                mode += 1

        for b in range(len(colour)):
            if colour[b] == "":
                colour[b] = "255"
            colour[b] = int(colour[b])

        return colour


def load_theme(colours, path):

    f = open(path, encoding="utf-8")
    content = f.readlines()

    for p in content:
        p = p.strip()
        if "# " in p:
            p = p.split("# ")[0]
        if not p:
            continue
        if p[0] == "#" and ("-" in p[:7] or " " in p[:7] or "\t" in p[:7]):
            continue
        if p.startswith("deco="):
            colours.deco = p.split("=", 1)[1].strip()
        if "light-mode" in p:
            colours.light_mode()
        if 'window frame' in p:
            colours.window_frame = get_colour_from_line(p)
        if 'gallery highlight' in p:
            colours.gallery_highlight = get_colour_from_line(p)
        if 'index playing' in p:
            colours.index_playing = get_colour_from_line(p)
        if 'time playing' in p:
            colours.time_text = get_colour_from_line(p)
        if 'artist playing' in p:
            colours.artist_playing = get_colour_from_line(p)
        if 'album line' in p:  # Bad name
            colours.album_text = get_colour_from_line(p)
        if 'track album' in p:
            colours.album_text = get_colour_from_line(p)
        if 'album playing' in p:
            colours.album_playing = get_colour_from_line(p)
        if 'top panel' in p or 'player background' in p:
            colours.top_panel_background = get_colour_from_line(p)

            colours.status_text_over = rgb_add_hls(colours.top_panel_background, 0, 0.83, 0)
            colours.status_text_normal = rgb_add_hls(colours.top_panel_background, 0, 0.30, -0.15)

            if test_lumi(colours.bottom_panel_colour) < 0.2:
                colours.corner_icon = [0, 0, 0, 60]
            elif test_lumi(colours.bottom_panel_colour) < 0.8:
                colours.corner_icon = [40, 40, 40, 255]
            else:
                colours.corner_icon = [255, 255, 255, 30]

            if test_lumi(colours.bottom_panel_colour) < 0.2:
                colours.corner_icon = [0, 0, 0, 60]

            if not colours.lm:
                colours.corner_button = rgb_add_hls(colours.top_panel_background, 0, 0.18, 0)

        if 'corner button off' in p:
            colours.corner_button = get_colour_from_line(p)
        if 'corner button on' in p:
            colours.corner_button_active = get_colour_from_line(p)
        if 'menu button normal' in p:
            colours.status_text_normal = get_colour_from_line(p)
        if 'menu button hover' in p:
            colours.status_text_over = get_colour_from_line(p)
        if 'queue panel' in p:
            colours.queue_background = get_colour_from_line(p)
        if 'side panel' in p:
            colours.side_panel_background = get_colour_from_line(p)
            colours.playlist_box_background = colours.side_panel_background
        if 'gallery background' in p:
            colours.gallery_background = get_colour_from_line(p)
        if 'playlist panel' in p:  # bad name
            colours.playlist_panel_background = get_colour_from_line(p)
        if 'tracklist panel' in p:
            colours.playlist_panel_background = get_colour_from_line(p)
        if 'track line' in p:
            colours.title_text = get_colour_from_line(p)
        if 'track missing' in p:
            colours.playlist_text_missing = get_colour_from_line(p)
        if 'playing highlight' in p:
            colours.row_playing_highlight = get_colour_from_line(p)
        if 'track time' in p:
            colours.bar_time = get_colour_from_line(p)
        if 'fav line' in p:
            colours.star_line = get_colour_from_line(p)
        if 'folder title' in p:
            colours.folder_title = get_colour_from_line(p)
        if 'folder line' in p:
            colours.folder_line = get_colour_from_line(p)
        if 'buttons off' in p:
            colours.media_buttons_off = get_colour_from_line(p)
        if 'buttons over' in p:
            colours.media_buttons_over = get_colour_from_line(p)
        if 'buttons active' in p:
            colours.media_buttons_active = get_colour_from_line(p)
        if 'playing time' in p:
            colours.time_playing = get_colour_from_line(p)
        if 'track index' in p:
            colours.index_text = get_colour_from_line(p)
        if 'track playing' in p:
            colours.title_playing = get_colour_from_line(p)
        if 'select highlight' in p:
            colours.row_select_highlight = get_colour_from_line(p)
        if 'track artist' in p:
            colours.artist_text = get_colour_from_line(p)
        if 'tab active line' in p:  # bad name
            colours.tab_text_active = get_colour_from_line(p)
        if 'tab line' in p:  # bad name
            colours.tab_text = get_colour_from_line(p)
        if 'tab active text' in p:
            colours.tab_text_active = get_colour_from_line(p)
        if 'tab text' in p:
            colours.tab_text = get_colour_from_line(p)
        if 'tab background' in p:
            colours.tab_background = get_colour_from_line(p)
        if 'tab over' in p:
            colours.tab_highlight = get_colour_from_line(p)
        if 'tab active background' in p:
            colours.tab_background_active = get_colour_from_line(p)
        if 'title info' in p:
            colours.side_bar_line1 = get_colour_from_line(p)
        if 'extra info' in p:
            colours.side_bar_line2 = get_colour_from_line(p)
        if 'bottom title' in p:
            colours.bar_title_text = get_colour_from_line(p)
        if 'scroll bar' in p:
            colours.scroll_colour = get_colour_from_line(p)
        if 'seek bar' in p:
            colours.seek_bar_fill = get_colour_from_line(p)
        if 'seek bg' in p:
            colours.seek_bar_background = get_colour_from_line(p)
        if 'volume bar' in p:
            colours.volume_bar_fill = get_colour_from_line(p)
        if 'volume bg' in p:
            colours.volume_bar_background = get_colour_from_line(p)
        if 'mode off' in p:
            colours.mode_button_off = get_colour_from_line(p)
        if 'mode over' in p:
            colours.mode_button_over = get_colour_from_line(p)
        if 'mode on' in p:
            colours.mode_button_active = get_colour_from_line(p)
        if 'art border' in p:
            colours.art_box = get_colour_from_line(p)
        if 'tb line' in p:
            colours.tb_line = get_colour_from_line(p)
        if 'music vis' in p:
            colours.vis_colour = get_colour_from_line(p)
        if 'menu background' in p:
            colours.menu_background = get_colour_from_line(p)
        if 'menu text' in p:
            colours.menu_text = get_colour_from_line(p)
        if 'menu disable' in p:
            colours.menu_text_disabled = get_colour_from_line(p)
        if 'menu icons' in p:
            colours.menu_icons = get_colour_from_line(p)
        if 'menu highlight' in p:
            colours.menu_highlight_background = get_colour_from_line(p)
        if 'menu border' in p:
            colours.menu_tab = get_colour_from_line(p)
        if 'lyrics showcase' in p:
            colours.lyrics = get_colour_from_line(p)
        if 'bottom panel' in p:
            colours.bottom_panel_colour = get_colour_from_line(p)
            # colours.menu_background = colours.bottom_panel_colour
        if 'mini bg' in p:
            colours.mini_mode_background = get_colour_from_line(p)
        if 'mini border' in p:
            colours.mini_mode_border = get_colour_from_line(p)
        if 'mini text 1' in p:
            colours.mini_mode_text_1 = get_colour_from_line(p)
        if 'mini text 2' in p:
            colours.mini_mode_text_2 = get_colour_from_line(p)
        if 'column-' in p:
            key = p[p.find("column-") + 7:].replace("-", " ").lower().title().rstrip()
            value = get_colour_from_line(p)
            colours.column_colours[key] = value
        if 'column+' in p:
            key = p[p.find("column+") + 7:].replace("-", " ").lower().title().rstrip()
            value = get_colour_from_line(p)
            colours.column_colours_playing[key] = value
        if 'menu bg' in p:
            colours.menu_background = get_colour_from_line(p)
        if 'playlist box bg' in p:  # bad name
            colours.playlist_box_background = get_colour_from_line(p)
        if 'playlist background' in p:
            colours.playlist_box_background = get_colour_from_line(p)

        if 'box background' in p:
            colours.box_background = get_colour_from_line(p)
        if 'box border' in p:
            colours.box_border = get_colour_from_line(p)
        if 'box text border' in p:
            colours.box_text_border = get_colour_from_line(p)
        if 'box text label' in p:
            colours.box_text_label = get_colour_from_line(p)

        if 'box title text' in p:
            colours.box_title_text = get_colour_from_line(p)
        if 'box text normal' in p:
            colours.box_text = get_colour_from_line(p)
        if 'box sub text' in p:
            colours.box_sub_text = get_colour_from_line(p)
        if 'box input text' in p:
            colours.box_input_text = get_colour_from_line(p)

        if 'box button text highlight' in p:
            colours.box_button_text_highlight = get_colour_from_line(p)
        if 'box button text normal' in p:
            colours.box_button_text = get_colour_from_line(p)
        if 'box button background normal' in p:
            colours.box_button_background = get_colour_from_line(p)
        if 'box button background highlight' in p:
            colours.box_button_background_highlight = get_colour_from_line(p)
        if 'box button border' in p:
            colours.box_check_border = get_colour_from_line(p)

        if "window buttons background" in p:
            colours.window_buttons_bg = get_colour_from_line(p)
        if "window buttons on" in p:
            colours.window_buttons_bg_over = get_colour_from_line(p)
        if "window buttons icon off" in p:
            colours.window_button_icon_off = get_colour_from_line(p)
            colours.window_button_x_off = colours.window_button_icon_off
        if "window buttons icon over" in p:
            colours.window_buttons_icon_over = get_colour_from_line(p)
            colours.window_button_x_on = colours.window_buttons_icon_over
        if "window button x on" in p:
            colours.window_button_x_on = get_colour_from_line(p)
        if "window button x off" in p:
            colours.window_button_x_off = get_colour_from_line(p)
        if "column bar background" in p:
            colours.column_bar_background = get_colour_from_line(p)

        if "artist bio background" in p:
            colours.artist_bio_background = get_colour_from_line(p)
        if "artist bio text" in p:
            colours.artist_bio_text = get_colour_from_line(p)
        # if "panel button off" in p:
        #     colours.corner_button = get_colour_from_line(p)
        # if "panel button on" in p:
        #     colours.corner_button_active = get_colour_from_line(p)

    colours.post_config()
    if colours.lm:
        colours.light_mode()


class Drawable:
    def __int__(self):
        self.location = 1
        self.x = 0
        self.y = 0
        self.w = 100
        self.y = 100
        self.rect = None
        self.texture = None

class Deco:
    def __init__(self, tauon):
        self.tauon = tauon
        self.renderer = None
        self.pctl = tauon.pctl
        self.prefs = tauon.prefs
        self.drawables = []

    def unload(self):
        for item in self.drawables:
            SDL_DestroyTexture(item.texture)
        self.drawables.clear()

    def load(self, name):
        self.unload()
        if not name:
            return

        decos = self.get_themes(deco=True)

        if name not in decos:
            print("Missing deco file")
            return

        target = decos[name]
        with open(target) as f:
            j = json.load(f)

        if "images" not in j:
            return
        images = j["images"]
        if not images:
            return
        item = images[0]
        if item.get("location") != 1:
            return
        x = int(item.get("x-margin", 0))
        y = int(item.get("y-margin", 0))
        opacity = int(item.get("opacity", 100))
        logical_h = int(item.get("logical-height", 400))
        file = item.get("file")
        if file:
            path = os.path.join(os.path.dirname(target), file)
            im = Image.open(path)
        else:
            s = item.get("image-data")
            if not s:
                return
            b = io.BytesIO(base64.b64decode(s))
            im = Image.open(b)

        w, h = im.size
        scale = self.tauon.gui.scale
        if not abs(h - (logical_h * scale)) < 10:
            new_h = round(logical_h * scale)
            ratio = w / h
            new_w = round(new_h * ratio)
            im = im.resize((new_w, new_h), Image.Resampling.LANCZOS)
            w, h = new_w, new_h

        g = io.BytesIO()
        g.seek(0)
        im.save(g, 'PNG')
        g.seek(0)
        wop = rw_from_object(g)
        s_image = IMG_Load_RW(wop, 0)
        texture = SDL_CreateTextureFromSurface(self.renderer, s_image)
        SDL_SetTextureAlphaMod(texture, opacity)
        SDL_FreeSurface(s_image)
        sdl_rect = SDL_Rect(0, 0, w, h)

        drawable = Drawable()
        drawable.x = x
        drawable.y = y
        drawable.w = w
        drawable.h = h
        drawable.rect = sdl_rect
        drawable.texture = texture
        self.drawables.append(drawable)

    def draw(self, ddt, x, y, pretty_text=False):
        if self.drawables:
            d = self.drawables[0]
            d.rect.x = round(x - int(d.w + round(d.x * self.tauon.gui.scale)))
            d.rect.y = round(y - int(d.h + round(d.y * self.tauon.gui.scale)))
            SDL_RenderCopy(self.renderer, d.texture, None, d.rect)

            if pretty_text:
                ddt.pretty_rect = (d.rect.x, d.rect.y, d.rect.w, d.rect.h)
                ddt.alpha_bg = True




# Tauon Music Box - Web interface module

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
import sys
if sys.platform != 'win32':
    import fcntl
import html
import time
import random
import json
import io
import hashlib
import os
import subprocess
from http.server import HTTPServer
from http.server import BaseHTTPRequestHandler
from socketserver import ThreadingMixIn
from t_modules.t_extra import Timer


def send_file(path, mime, server):
    range_req = False
    start = 0
    end = 0

    if "Range" in server.headers:
        range_req = True
        b = server.headers["Range"].split("=")[1]
        start, end = b.split("-")
        start = int(start)

    with open(path, "rb") as f:

        f.seek(0, 2)
        length = f.tell()
        f.seek(start, 0)

        if range_req:
            server.send_response(206)
            server.send_header("Content-Range", f"bytes {start}-{length - 1}/{length}")
            server.send_header("Content-Length", str(length))
            server.send_header("Content-Type", mime)
            f.seek(start)

        else:
            server.send_response(200)
            server.send_header("Accept-Ranges", "bytes")
            server.send_header("Content-Type", mime)
            server.send_header("Content-Length", str(length))

        server.end_headers()

        while True:
            data = f.read(1000)
            if not data:
                break
            server.wfile.write(data)

def webserve(pctl, prefs, gui, album_art_gen, install_directory, strings, tauon):

    if prefs.enable_web is False:
        return 0

    gui.web_running = True

    class Server(BaseHTTPRequestHandler):

        def log_message(self, format, *args):
            tauon.gui.console.print(format % args)
            return

        def send_file(self, path, mime):
            self.send_response(200)
            self.send_header("Content-type", mime)
            self.end_headers()

            with open(path, "rb") as f:
                self.wfile.write(f.read())

        def get_track_id(self, track):
            return hashlib.md5((str(track.index) + track.title + track.artist).encode()).hexdigest()

        def do_GET(self):

            path = self.path

            # print(self.headers)
            # print(path)

            if path == "/listenalong/":
                self.send_response(302)
                self.send_header('Location', "/listenalong")
                self.end_headers()

            elif path == "/listenalong":
                self.send_file(install_directory + "/templates/radio.html", "text/html")
            elif path == "/favicon.png":
                self.send_file(install_directory + "/assets/icon-64.png", 'image/png')
            elif path == "/listenalong/play.svg":
                self.send_file(install_directory + "/templates/play.svg", 'image/svg+xml')
            elif path == "/listenalong/pause.svg":
                self.send_file(install_directory + "/templates/pause.svg", 'image/svg+xml')
            elif path == "/listenalong/stop.svg":
                self.send_file(install_directory + "/templates/stop.svg", 'image/svg+xml')
            elif path == "/radio/radio.js":
                self.send_file(install_directory + "/templates/radio.js", "application/javascript")
            elif path == "/radio/theme.css":
                self.send_file(install_directory + "/templates/theme.css", "text/css")
            elif path == "/radio/logo-bg.png":
                self.send_file(install_directory + "/templates/logo-bg.png", 'image/png')

            elif path.startswith("/llapi/audiofile/"):
                value = path[17:]
                track = pctl.playing_object()
                if not track:
                    self.send_response(403)
                    self.end_headers()
                    return
                sid = self.get_track_id(track)
                if sid != value or track.is_network:
                    self.send_response(403)
                    self.end_headers()
                    return
                if not track.fullpath or not os.path.isfile(track.fullpath):
                    self.send_response(404)
                    self.end_headers()
                    return
                mime = "audio/mpeg"
                if track.file_ext == "FLAC":
                    mime = "audio/flac"
                if track.file_ext == "OGG" or track.file_ext == "OPUS" or track.file_ext == "OGA":
                    mime = "audio/ogg"
                if track.file_ext == "M4A":
                    mime = "audio/mp4"
                send_file(track.fullpath, mime, self)

            elif path == "/llapi/poll":
                self.send_response(200)
                self.send_header("Content-type", "application/json")

                ip = self.client_address[0]
                timer = tauon.listen_alongers.get(ip)
                if not timer:
                    tauon.listen_alongers[ip] = Timer()
                else:
                    timer.set()

                track = pctl.playing_object()
                if track is None:
                    data = {"status": 0}
                else:
                    data = {"status": pctl.playing_state,
                            "id": self.get_track_id(track),
                            "position": pctl.playing_time,
                            "duration": track.length,
                            "title": track.title,
                            "artist": track.artist,
                            }
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path.startswith("/llapi/picture/"):
                value = path[15:]
                track = pctl.playing_object()
                if not track:
                    self.send_response(403)
                    self.end_headers()
                    return
                sid = self.get_track_id(track)
                if sid != value or track.is_network:
                    self.send_response(403)
                    self.end_headers()
                    return

                try:
                    base64 = album_art_gen.get_base64(track, (300, 300)).decode()
                    data = {"image_data": base64}
                except:
                    #raise
                    data = {"image_data": "None"}

                lyrics = tauon.synced_to_static_lyrics.get(track)
                lyrics = html.escape(lyrics).replace("\r\n", "\n").replace("\r", "\n").replace("\n", "<br>")
                data["lyrics"] = lyrics

                data = json.dumps(data).encode()
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.send_header("Content-length", str(len(data)))
                self.end_headers()
                self.wfile.write(data)

            else:
                self.send_response(404)
                self.end_headers()
                self.wfile.write(b"404 Not found")

    class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
        pass

    try:
        httpd = ThreadedHTTPServer(("0.0.0.0", prefs.metadata_page_port), Server)
        tauon.radio_server = httpd
        httpd.serve_forever()
        httpd.server_close()
    except OSError:
        print("Not starting radio page server, already running?")


def webserve2(pctl, prefs, gui, album_art_gen, install_directory, strings, tauon):

    play_timer = Timer()

    class Server(BaseHTTPRequestHandler):

        def log_message(self, format, *args):
            tauon.gui.console.print(format % args)
            return

        def run_command(self, callback):
            self.send_response(200)
            #self.send_header("Content-type", "application/json")
            self.end_headers()
            callback()
            self.wfile.write(b"OK")

        def parse_trail(self, text):

            params = {}
            both = text.split("?")
            levels = both[0].split("/")
            if len(both) > 1:
                pairs = both[2].split("&")
                for p in pairs:
                    aa, bb = p.split("=")
                    params[aa] = bb

            return levels, params

        def get_track(self, track_position, playlist_index=None, track=None, album_id=-1):
            if track is None:
                if playlist_index is None:
                    playlist = pctl.multi_playlist[pctl.active_playlist_playing][2]
                else:
                    playlist = pctl.multi_playlist[playlist_index][2]
                track_id = playlist[track_position]
                track = pctl.g(track_id)

            data = {}
            data["title"] = track.title
            data["artist"] = track.artist
            data["album"] = track.album
            data["album_artist"] = track.album_artist
            if not track.album_artist:
                data["album_artist"] = track.artist
            data["duration"] = int(track.length * 1000)
            data["id"] = track.index
            data["position"] = track_position
            data["path"] = track.fullpath
            data["album_id"] = album_id
            data["has_lyrics"] = track.lyrics != ""
            data["track_number"] = str(track.track_number).lstrip("0")
            data["can_download"] = not track.is_cue and not track.is_network

            return data

        def do_GET(self):

            path = self.path
            #print(self.headers)
            if tauon.remote_limited and not tauon.chrome_mode:
                self.send_response(404)
                self.end_headers()
                self.wfile.write(b"404 Not found")
                return
            if tauon.remote_limited and (not path.startswith("/api1/pic/medium/") and not path.startswith("/api1/file/")):
                self.send_response(404)
                self.end_headers()
                self.wfile.write(b"404 Not found")
                return

            if path.startswith("/api1/pic/small/"):
                value = path[16:]
                if value.isalnum() and int(value) in pctl.master_library:
                    track = pctl.g(int(value))
                    raw = album_art_gen.save_thumb(track, (75, 75), "")
                    if raw:
                        self.send_response(200)
                        self.send_header("Content-type", "image/jpg")
                        self.end_headers()
                        self.wfile.write(raw.read())
                    else:
                        self.send_response(404)
                        self.end_headers()
                        self.wfile.write(b"No image found")

                else:
                    self.send_response(404)
                    self.end_headers()
                    self.wfile.write(b"Invalid parameter")

            if path.startswith("/api1/pic/medium/"):
                value = path[17:]
                print(value)
                if value.isalnum() and int(value) in pctl.master_library:
                    track = pctl.g(int(value))
                    raw = album_art_gen.save_thumb(track, (1000, 1000), "")
                    if raw:
                        self.send_response(200)
                        self.send_header("Content-type", "image/jpg")
                        self.end_headers()
                        self.wfile.write(raw.read())
                    else:
                        self.send_response(404)
                        self.end_headers()
                        self.wfile.write(b"No image found")

                else:
                    self.send_response(404)
                    self.end_headers()
                    self.wfile.write(b"Invalid parameter")

            if path.startswith("/api1/lyrics/"):
                value = path[13:]
                if value.isalnum() and int(value) in pctl.master_library:
                    track = pctl.g(int(value))
                    data = {}
                    data["track_id"] = track.index
                    data["lyrics_text"] = track.lyrics

                    self.send_response(200)
                    self.send_header("Content-type", "application/json")
                    self.end_headers()
                    data = json.dumps(data).encode()
                    self.wfile.write(data)
                else:
                    self.send_response(404)
                    self.end_headers()
                    self.wfile.write(b"Invalid parameter")

            # elif path.startswith("/api1/stream/"):
            #     param = path[13:]
            #
            #     if param.isdigit() and int(param) in pctl.master_library:
            #         track = pctl.master_library[int(param)]
            #         mime = "audio/mpeg"
            #         #mime = "audio/ogg"
            #         self.send_response(200)
            #         self.send_header("Content-type", mime)
            #         self.end_headers()
            #
            #         cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libopus", "-f", "ogg", "-"]
            #         #cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libvorbis", "-f", "ogg", "-"]
            #         #cmd = ["ffmpeg", "-i", track.fullpath, "-c:a", "libmp3lame", "-f", "mp3", "-"]
            #         encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            #         while True:
            #             data = encoder.stdout.read(1024)
            #             if data:
            #                 self.wfile.write(data)
            elif path.startswith("/api1/playinghit/"):
                param = path[17:]
                if param.isdigit() and int(param) in pctl.master_library:
                    t = play_timer.hit()
                    if 0 < t < 5:
                        tauon.star_store.add(int(param), t)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/file/"):
                param = path[11:]

                #print(self.headers)
                play_timer.hit()

                if param.isdigit() and int(param) in pctl.master_library:
                    track = pctl.master_library[int(param)]
                    mime = "audio/mpeg"
                    if track.file_ext == "FLAC":
                        mime = "audio/flac"
                    if track.file_ext == "OGG" or track.file_ext == "OPUS" or track.file_ext == "OGA":
                        mime = "audio/ogg"
                    if track.file_ext == "M4A":
                        mime = "audio/mp4"
                    send_file(track.fullpath, mime, self)

            elif path.startswith("/api1/start/"):

                levels, _ = self.parse_trail(path)
                if len(levels) == 5:
                    playlist = levels[3]
                    position = levels[4]
                    if playlist.isdigit() and position.isdigit():
                        position = int(position)
                        playlist = int(playlist)
                        pl = tauon.id_to_pl(int(playlist))
                        if pl is not None and pl < len(pctl.multi_playlist):
                            playlist = pctl.multi_playlist[pl][2]
                            if position < len(playlist):
                                tauon.switch_playlist(pl, cycle=False, quiet=True)
                                pctl.jump(playlist[position], position)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path == "/api1/play":
                self.run_command(tauon.pctl.play)
            elif path == "/api1/pause":
                self.run_command(tauon.pctl.pause_only)
            elif path == "/api1/next":
                self.run_command(tauon.pctl.advance)
            elif path == "/api1/back":
                self.run_command(tauon.pctl.back)
            elif path == "/api1/shuffle":
                self.run_command(tauon.toggle_random)
            elif path == "/api1/repeat":
                self.run_command(tauon.toggle_repeat)
            elif path == "/api1/auto-stop":
                tauon.pctl.auto_stop ^= True
                tauon.gui.update += 1
            elif path == "/api1/version":
                data = {"version": 1}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path == "/api1/playlists":

                l = []
                for item in pctl.multi_playlist:
                    p = {}
                    p["name"] = item[0]
                    p["id"] = str(item[6])
                    p["count"] = len(item[2])
                    l.append(p)
                data = {"playlists": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path.startswith("/api1/albumtracks/"):
                # Get tracks that appear in an album /albumtracks/plid/albumid
                levels, _ = self.parse_trail(path)
                l = []
                if len(levels) == 5 and levels[3].isdigit() and levels[4].isdigit():
                    pl = tauon.id_to_pl(int(levels[3]))
                    if pl is not None:
                        _, album, _ = tauon.get_album_info(int(levels[4]), pl)
                        # print(album)
                        for p in album:
                            l.append(self.get_track(p, pl, album_id=int(levels[4])))

                data = {"tracks": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path.startswith("/api1/trackposition/"):
                # get track /trackposition/plid/playlistposition
                levels, _ = self.parse_trail(path)
                if len(levels) == 5 and levels[3].isdigit() and levels[4].isdigit():
                    pl = tauon.id_to_pl(int(levels[3]))
                    if pl is not None:

                        data = self.get_track(int(levels[4]), pl)

                        playlist = pctl.multi_playlist[pl][2]
                        p = int(levels[4])
                        if p < len(playlist):
                            track = pctl.g(playlist[p])
                            while True:
                                if p < 0 or pctl.g(playlist[p]).parent_folder_path != track.parent_folder_path:
                                    p += 1
                                    break
                                p -= 1
                            data["album_id"] = p

                            self.send_response(200)
                            self.send_header("Content-type", "application/json")
                            self.end_headers()
                            data = json.dumps(data).encode()
                            self.wfile.write(data)
                        else:
                            self.send_response(404)
                            self.send_header("Content-type", "text/plain")
                            self.end_headers()
                            self.wfile.write(b"404 invalid track position")
                    else:
                        self.send_response(404)
                        self.send_header("Content-type", "text/plain")
                        self.end_headers()
                        self.wfile.write(b"404 playlist not found")
                else:
                    self.send_response(404)
                    self.send_header("Content-type", "text/plain")
                    self.end_headers()
                    self.wfile.write(b"404 invalid track")

            elif path.startswith("/api1/setvolume/"):
                key = path[16:]
                if key.isdigit():
                    volume = int(key)
                    volume = max(volume, 0)
                    volume = min(volume, 100)
                    pctl.player_volume = volume
                    pctl.set_volume()

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/setvolumerel/"):
                key = path[19:]
                if key.lstrip("-").isdigit() and key.count("-") < 2:
                    volume = pctl.player_volume + int(key)
                    volume = max(volume, 0)
                    volume = min(volume, 100)
                    pctl.player_volume = volume
                    pctl.set_volume()

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/seek1k/"):
                key = path[13:]
                if key.isdigit():
                    pctl.seek_decimal(int(key) / 1000)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/seek/"):
                key = path[11:]
                if key.isdigit():
                    pctl.seek_time(int(key) / 1000)

                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                self.wfile.write(b"OK")

            elif path.startswith("/api1/tracklist/"):
                # Return all tracks in a playlist /tracklist/plid
                key = path[16:]
                l = []
                if key.isdigit():
                    pl = tauon.id_to_pl(int(key))
                    if pl is not None and pl < len(pctl.multi_playlist):
                        playlist = pctl.multi_playlist[pl][2]
                        parent = ""
                        album_id = 0
                        for i, id in enumerate(playlist):
                            tr = pctl.g(id)
                            if i == 0:
                                parent = tr.parent_folder_path
                            elif parent != tr.parent_folder_path:
                                parent = tr.parent_folder_path
                                album_id = i

                            l.append(self.get_track(i, pl, album_id=album_id))

                data = {"tracks": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path.startswith("/api1/albums/"):
                # Returns lists of tracks that are start of albums /albums/plid
                key = path[13:]
                l = []
                if key.isdigit():
                    pl = tauon.id_to_pl(int(key))
                    if pl is not None:
                        dex = tauon.reload_albums(True, pl)
                        # print(dex)
                        for a in dex:
                            l.append(self.get_track(a, pl, album_id=a))

                data = {"albums": l}
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = json.dumps(data).encode()
                self.wfile.write(data)

            elif path == "/api1/status":
                self.send_response(200)
                self.send_header("Content-type", "application/json")
                self.end_headers()
                data = {
                    "status": "stopped",
                    "inc": pctl.db_inc,
                    "shuffle": pctl.random_mode == True,
                    "repeat": pctl.repeat_mode == True,
                    "progress": 0,
                    "auto_stop": tauon.pctl.auto_stop == True,
                    "volume": pctl.player_volume,
                    "playlist": str(tauon.get_playing_playlist_id()),
                    "playlist_length": len(pctl.multi_playlist[pctl.active_playlist_playing][2])
                }
                if pctl.playing_state == 1:
                    data["status"] = "playing"
                if pctl.playing_state == 2:
                    data["status"] = "paused"
                track = pctl.playing_object()
                if track:
                    data["id"] = track.index
                    data["title"] = track.title
                    data["artist"] = track.artist
                    data["album"] = track.album
                    data["progress"] = int(round(pctl.playing_time * 1000))
                    data["track"] = self.get_track(0, 0, track)

                p = pctl.playlist_playing_position
                data["position"] = p
                data["album_id"] = 0
                playlist = pctl.playing_playlist()

                if p < len(playlist):
                    while True:
                        if p < 0 or pctl.g(playlist[p]).parent_folder_path != track.parent_folder_path:
                            p += 1
                            break
                        p -= 1
                    data["album_id"] = p

                data = json.dumps(data).encode()
                self.wfile.write(data)

            else:
                self.send_response(404)
                self.end_headers()
                self.wfile.write(b"404 Not found")
            tauon.wake()

    class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
        pass

    try:
        httpd = ThreadedHTTPServer(("0.0.0.0", 7814), Server)
        httpd.serve_forever()
        httpd.server_close()
    except OSError:
        print("Not starting web api server, already running?")


def controller(tauon):
    import base64
    class Server(BaseHTTPRequestHandler):

        def log_message(self, format, *args):
            tauon.gui.console.print(format % args)
            return

        def do_GET(self):

            path = self.path
            if path == "/raise/":
                tauon.raise_window()
            if path == "/playpause/":
                if tauon.pctl.playing_state == 0:
                    tauon.pctl.play()
                else:
                    tauon.pctl.pause()
            if path == "/play/":
                tauon.pctl.play()
            if path == "/pause/":
                tauon.pctl.pause_only()
            if path == "/stop/":
                tauon.pctl.stop()
            if path == "/next/":
                tauon.pctl.advance()
            if path == "/previous/":
                tauon.pctl.back()
            if path == "/shuffle/":
                tauon.toggle_random()
            if path == "/repeat/":
                tauon.toggle_repeat()
            if path.startswith("/open/"):
                rest = path[6:]
                path = base64.urlsafe_b64decode(rest.encode()).decode()
                tauon.open_uri(path)

            self.send_response(200)
            self.end_headers()
            tauon.wake()
    print("Start controller server")
    try:
        httpd = HTTPServer(("127.0.0.1", 7813), Server)
        httpd.serve_forever()
        httpd.server_close()
    except OSError:
        print("Not starting controller server, already running?")


def authserve(tauon):

    class Server(BaseHTTPRequestHandler):

        def log_message(self, format, *args):
            tauon.gui.console.print(format % args)
            return

        def do_GET(self):
            code = ""
            path = self.path
            if path.startswith("/spotredir"):
                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                code = path.split("code=")
                if len(code) > 1:
                    code = code[1]
                    self.wfile.write(b"You can close this now and return to Tauon Music Box")
                tauon.wake()

            else:
                self.send_response(400)
                self.end_headers()

            if code:
                tauon.spot_ctl.paste_code(code)

    httpd = HTTPServer(("127.0.0.1", 7811), Server)
    httpd.serve_forever()
    httpd.server_close()

import struct
class VorbisMonitor():

    def __init__(self):

        self.tauon = None
        self.reset()
        self.enable = True
        self.synced = False
        self.buffer = io.BytesIO()
        self.tries = 0

    def reset(self, tries=0):
        self.enable = True
        self.synced = False
        self.buffer = io.BytesIO()
        self.tries = tries

    def input(self, data):

        if not self.enable:
            return

        b = self.buffer
        b.seek(0, io.SEEK_END)
        b.write(data)

        # Check theres enough data to decode header
        b.seek(0, io.SEEK_END)
        l = b.tell()
        if l < 128:
            print("Not enough data to parse vorbis")
            return

        # Get page length
        b.seek(0, io.SEEK_SET)
        ogg = b.read(4)

        if not ogg == b"OggS":
            f = data.find(b"Oggs")
            self.reset(self.tries)
            if f > -1:
                print("Ogg stream synced")
                data = data[f:]
                b = self.buffer
                b.write(data)
            else:
                self.tries += 1
                if self.tries > 100:
                    print("Giving up looking for OGG pages")
                    self.enable = False
                return

            # self.enable = False
            # return

        b.seek(0, io.SEEK_SET)
        header = struct.unpack('<4sBBqIIiB', b.read(27))
        segs = struct.unpack('B' * header[7], b.read(header[7]))

        length = 0
        for s in segs:
            length += s

        length += 27 + header[7]

        if l > length:
            h = b.read(7)
            # print(h)
            if h == b"\x03vorbis" or h == b"OpusTag":
                if h == b"OpusTag":
                    b.seek(1, io.SEEK_CUR)

                vendor_length = int.from_bytes(b.read(4), byteorder='little')
                vendor = int.from_bytes(b.read(vendor_length), byteorder='little')
                comment_list_length = int.from_bytes(b.read(4), byteorder='little')

                found_tags = {}
                for i in range(comment_list_length):
                    comment_length = int.from_bytes(b.read(4), byteorder='little')
                    comment = b.read(comment_length)

                    key, value = comment.decode().split("=", 1)

                    if key == "title":
                        found_tags["title"] = value
                    if key == "artist":
                        found_tags["artist"] = value
                    if key == "year":
                        found_tags["year"] = value
                    if key == "album":
                        found_tags["album"] = value

                line = ""
                if "title" in found_tags:
                    line += found_tags["title"]
                    if "artist" in found_tags:
                        line = found_tags["artist"] + " - " + line

                if self.tauon.radiobox.parse_vorbis_okay():
                    self.tauon.pctl.found_tags = found_tags
                    self.tauon.pctl.tag_meta = line

                print("Found vorbis comment")
                print(line)

            # Consume page from buffer
            b.seek(length, io.SEEK_SET)
            new = io.BytesIO()
            new.write(b.read())
            self.buffer = new

vb = VorbisMonitor()

def stream_proxy(tauon):

    class Server(BaseHTTPRequestHandler):
        def log_message(self, format, *args):
            tauon.gui.console.print(format % args)
            return

        def do_GET(self):

            self.send_response(200)
            self.send_header("Content-type", "audio/ogg")
            self.end_headers()

            position = 0
            vb.reset()
            vb.tauon = tauon

            while True:
                if not tauon.stream_proxy.download_running:
                    return

                while position < tauon.stream_proxy.c:
                    if position not in tauon.stream_proxy.chunks:
                        print("The buffer was deleted too soon!")
                        return

                    self.wfile.write(tauon.stream_proxy.chunks[position])

                    if tauon.prefs.backend == 4:
                        vb.input(tauon.stream_proxy.chunks[position])

                    position += 1

                else:
                    time.sleep(0.01)

    httpd = HTTPServer(("127.0.0.1", 7812), Server)
    httpd.serve_forever()
    httpd.server_close()




# Tauon Music Box - URL stream download and encoding module

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
import mimetypes
import urllib.request
import threading
import time
import subprocess
import os
import sys
if sys.platform != 'win32':
    import fcntl
import datetime
import io
import shutil
import mutagen
import copy
from PIL import Image
from mutagen.flac import Picture
import base64
from t_modules.t_extra import filename_safe
from t_modules.t_webserve import vb

class StreamEnc:

    def __init__(self, tauon):
        self.tauon = tauon
        self.download_running = False
        self.encode_running = False
        self.pump_running = False
        self.feed_running = False

        self.download_process = False
        self.abort = False

        self.s_name = ""
        self.s_bitrate = ""
        self.s_genre = ""
        self.s_description = ""
        self.s_mime = ""
        self.s_format = ""

        self.chunks = {}
        self.c = 0
        self.url = None

    def stop(self):

        try:
            self.tauon.radiobox.websocket.close()
            print("Websocket closed")
        except:
            print("No socket to close?")

        self.abort = True
        self.tauon.radiobox.loaded_url = None

    def start_download(self, url):

        self.abort = True
        while self.download_running:
            time.sleep(0.01)
        while self.encode_running:
            time.sleep(0.01)
        while self.pump_running:
            time.sleep(0.01)

        self.__init__(self.tauon)

        self.url = url
        result = self.start_request()
        if not result:
            return False

        self.download_process = threading.Thread(target=self.pump)
        self.download_process.daemon = True
        self.download_process.start()
        return True


    def start_request(self):
        url = self.url
        def NiceToICY(self):
            class InterceptedHTTPResponse:
                pass
            if not url.endswith(".ts"):
                line = self.fp.readline().replace(b"ICY 200 OK\r\n", b"HTTP/1.0 200 OK\r\n")
            else:
                line = self.fp.readline()
            InterceptedSelf = InterceptedHTTPResponse()
            InterceptedSelf.fp = io.BufferedReader(io.BytesIO(line))
            InterceptedSelf.debuglevel = self.debuglevel
            InterceptedSelf._close_conn = self._close_conn
            return ORIGINAL_HTTP_CLIENT_READ_STATUS(InterceptedSelf)

        ORIGINAL_HTTP_CLIENT_READ_STATUS = urllib.request.http.client.HTTPResponse._read_status
        urllib.request.http.client.HTTPResponse._read_status = NiceToICY

        retry = 5
        while True:
            try:
                r = urllib.request.Request(self.url)
                #r.add_header('GET', '1')
                if not self.url.endswith(".ts"):
                    r.add_header('Icy-MetaData', '1')
                r.add_header('User-Agent', self.tauon.t_agent)
                print("Open URL.....")
                r = urllib.request.urlopen(r, timeout=20, cafile=self.tauon.ca)
                print("URL opened.")

            except Exception as e:

                print("URL error...")
                retry -= 1
                if retry > 0 and "Temporary" in str(e):
                    time.sleep(2)
                    print("RETRYING...")
                    continue
                else:
                    print("Connection failed")
                    print(str(e))
                    self.tauon.gui.show_message("Failed to establish a connection", str(e), mode="error")
                    return False
            break

        self.download_process = threading.Thread(target=self.run_download, args=([r]))
        self.download_process.daemon = True
        self.download_process.start()
        self.download_running = True
        return True

    def pump(self):
        aud = self.tauon.aud
        if self.tauon.prefs.backend != 4 or not aud:
            print("Radio error: Phazor not loaded")
            return
        self.pump_running = True

        rate = str(self.tauon.prefs.samplerate)
        cmd = [self.tauon.get_ffmpeg(), "-loglevel", "quiet", "-i", "pipe:0", "-acodec", "pcm_s16le", "-f", "s16le", "-ac", "2", "-ar",
               rate, "-"]

        startupinfo = None
        if self.tauon.msys:
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, startupinfo=startupinfo)
        if sys.platform != 'win32':
           fcntl.fcntl(decoder.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

        raw_audio = None
        max_read = int(10000)
        vb.reset()
        vb.tauon = self.tauon

        def feed(decoder):
            position = 0
            self.feed_running = True
            try:
                while True:
                    if position < self.tauon.stream_proxy.c:
                        if position not in self.tauon.stream_proxy.chunks:
                            print("The buffer was deleted too soon!")
                            break

                        chunk = self.chunks[position]
                        decoder.stdin.write(chunk)
                        vb.input(self.tauon.stream_proxy.chunks[position])
                        position += 1
                    else:
                        time.sleep(0.01)
                    if self.abort or not self.pump_running or not self.feed_running:
                        break
            except:
                self.feed_running = False
                raise
            print("Exit feeder")

        feeder = threading.Thread(target=feed, args=[decoder])
        feeder.daemon = True
        feeder.start()

        retry = 3

        while True:
            if not self.tauon.stream_proxy.download_running or self.abort:
                break
            if raw_audio is None:
                raw_audio = decoder.stdout.read(max_read)
            if raw_audio:
                r = aud.feed_ready(max_read)
                if r:
                    aud.feed_raw(len(raw_audio), raw_audio)
                    if len(raw_audio) < max_read:
                        time.sleep(0.01)
                    raw_audio = None
                    continue

            time.sleep(0.01)


        decoder.terminate()
        time.sleep(0.1)
        try:
            decoder.kill()
        except:
            pass

        self.pump_running = False


    def encode(self):

        self.encode_running = True

        try:

            while self.c < 20:
                if self.abort:
                    self.encode_running = False
                    return
                time.sleep(0.05)

            ext = ".opus"
            rate = "48000"
            codec = self.tauon.prefs.radio_record_codec.upper()
            if codec == "OGG":
                ext = ".ogg"
                rate = "44100"
            if codec == "MP3":
                ext = ".mp3"
                rate = "44100"
            if codec == "FLAC":
                ext = ".flac"
                rate = "44100"

            target_file = os.path.join(self.tauon.cache_directory, "stream" + ext)
            if os.path.isfile(target_file):
                os.remove(target_file)

            cmd = [self.tauon.get_ffmpeg(), "-loglevel", "quiet", "-i", "pipe:0", "-acodec", "pcm_s16le", "-f", "s16le", "-ac", "2", "-ar", rate, "-"]

            decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            if sys.platform != 'win32':
                fcntl.fcntl(decoder.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

            position = 0
            old_metadata = self.tauon.radiobox.song_key
            old_tags = self.tauon.pctl.found_tags

            ##cmd = ["opusenc", "--raw", "--raw-rate", "48000", "-", target_file]
            cmd = ["ffmpeg", "-loglevel", "quiet", "-f", "s16le", "-ar", rate, "-ac", "2", "-i", "pipe:0", target_file]
            encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

            def save_track():
                #self.tauon.recorded_songs.append(song)

                save_file = '{:%Y-%m-%d %H-%M-%S} - '.format(datetime.datetime.now())
                save_file += filename_safe(old_metadata)
                save_file = save_file.strip() + ext
                save_file = os.path.join(self.tauon.prefs.encoder_output, save_file)
                if os.path.exists(save_file):
                    os.remove(save_file)
                if not os.path.exists(self.tauon.prefs.encoder_output):
                    os.makedirs(self.tauon.prefs.encoder_output)
                shutil.move(target_file, save_file)

                # print(self.tauon.pctl.tag_history)
                # print(old_metadata)
                tags = self.tauon.pctl.tag_history.get(old_metadata, None)
                if tags:
                    print("Save metadata to file")
                    #print(tags)
                    muta = mutagen.File(save_file, easy=True)
                    muta["artist"] = tags.get("artist", "")
                    muta["title"] = tags.get("title", "")
                    muta["album"] = tags.get("album", "")
                    # if tags["image"]:
                    #     tags["image"].seek(0)
                    #     im = Image.open(tags["image"])
                    #     width, height = im.size
                    #     tags["image"].seek(0)
                    #
                    #     picture = Picture()
                    #     tags["image"].seek(0)
                    #     picture.data = tags["image"].read()
                    #     picture.type = 3
                    #     picture.desc = ""
                    #     picture.mime = "image/jpeg"
                    #     picture.width = width
                    #     picture.height = height
                    #
                    #     mode_to_bpp = {'1': 1, 'L': 8, 'P': 8, 'RGB': 24, 'RGBA': 32, 'CMYK': 32, 'YCbCr': 24, 'I': 32,
                    #                    'F': 32}
                    #     picture.depth = mode_to_bpp[im.mode]
                    #
                    #     picture_data = picture.write()
                    #     encoded_data = base64.b64encode(picture_data)
                    #     vcomment_value = encoded_data.decode("ascii")
                    #     muta["metadata_block_picture"] = [vcomment_value]

                    muta.save()

                target_pl = None
                for i, pl in enumerate(self.tauon.pctl.multi_playlist):
                    if pl[0] == "Saved Radio Tracks":
                        target_pl = i
                if target_pl is None:
                    self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title="Saved Radio Tracks"))
                    target_pl = len(self.tauon.pctl.multi_playlist) - 1

                load_order = self.tauon.pctl.LoadClass()
                load_order.playlist = self.tauon.pctl.multi_playlist[target_pl][6]
                load_order.target = save_file
                self.tauon.load_orders.append(copy.deepcopy(load_order))
                self.tauon.gui.update += 1

            while True:

                if self.abort:
                    decoder.terminate()
                    encoder.terminate()
                    time.sleep(0.1)
                    try:
                        decoder.kill()
                    except:
                        pass
                    try:
                        encoder.kill()
                    except:
                        pass

                    if os.path.exists(target_file):
                        if os.path.getsize(target_file) > 256000:
                            print("Save file")
                            save_track()
                        else:
                            print("Discard small file")
                            os.remove(target_file)

                    self.encode_running = False
                    self.tauon.pctl.tag_history.clear()
                    return

                if old_metadata != self.tauon.radiobox.song_key:
                    if self.c < 400 and not old_metadata:
                        old_metadata = self.tauon.radiobox.song_key
                    elif not os.path.exists(target_file) or os.path.getsize(target_file) < 100000:
                        old_metadata = self.tauon.radiobox.song_key
                    else:
                        print("Split and save file")
                        encoder.stdin.close()
                        try:
                            encoder.wait(timeout=4)
                        except:
                            pass
                        try:
                            encoder.kill()
                        except:
                            pass
                        if os.path.exists(target_file):
                            if os.path.getsize(target_file) > 256000:
                                save_track()
                            else:
                                print("Discard small file")
                                os.remove(target_file)
                        encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

                raw_audio = decoder.stdout.read(1000000)
                if raw_audio:
                    encoder.stdin.write(raw_audio)

                if position < self.c:
                    chunk = self.chunks[position]
                    position += 1
                    decoder.stdin.write(chunk)
                else:
                    time.sleep(0.005)

        except Exception as e:
            print("Encoder thread crashed!")
            print(str(e))
            #raise
            self.encode_running = False
            return

    def run_download(self, r):

        h = r.info()

        self.s_name = h.get("icy-name")
        metaint = h.get("icy-metaint")
        self.s_bitrate = h.get("icy-br")
        self.s_genre = h.get("icy-genre")
        self.s_description = h.get("icy-description")
        self.s_mime = h.get("Content-Type")

        print(self.s_mime)
        if self.s_mime == "audio/mpeg":
            self.s_format = "MP3"
        if self.s_mime == "audio/ogg":
            self.s_format = "OGG"
        if self.s_mime == "audio/aac":
            self.s_format = "AAC"
        if self.s_mime == "audio/aacp":
            self.s_format = "AAC+"

        test_done = 0

        icy = False
        m_remain = 0
        m = 0
        if metaint and int(metaint) > 0:
            m = int(metaint)
            m_remain = m
            icy = True

        maybe = b""

        if self.tauon.prefs.auto_rec:
            self.download_process = threading.Thread(target=self.encode)
            self.download_process.daemon = True
            self.download_process.start()

        try:
            while True:

                chunk = r.read(256)

                if self.abort:
                    r.close()
                    print("Abort stream connection")
                    self.download_running = False
                    return

                if chunk:
                    if not icy or m_remain > len(chunk):
                        # We're sure its data Its data, send it on
                        self.chunks[self.c] = chunk

                        # Delete old data
                        d = self.c - 90000
                        if d in self.chunks:
                            del self.chunks[d]

                        test_done += len(chunk)
                        self.c += 1
                        m_remain -= len(chunk)

                        continue
                    else:
                        # It may contain the metadata block, put it aside
                        maybe += chunk

                # Try to extract ICY tag
                if maybe:
                    data1 = maybe[:m_remain]
                    inter = maybe[m_remain:]

                    # Read the metadata length byte
                    if inter:
                        special = inter[0]
                        follow = special * 16

                        if len(inter) < follow + 2:
                            # Not enough data yet
                            continue

                        text = inter[1:follow + 1]
                        data2 = inter[follow + 1:]

                        self.chunks[self.c] = data1 + data2
                        # Delete old data
                        d = self.c - 90000
                        if d in self.chunks:
                            del self.chunks[d]

                        self.c += 1

                        test_done += len(data1)
                        test_done = 0

                        m_remain = m - len(data2)
                        test_done += len(data2)
                        maybe = b""

                        try:
                            meta = text.decode().rstrip("\x00")
                            for tag in meta.split(";"):
                                if '=' in tag:
                                    a, b = tag.split('=', 1)
                                    if a == 'StreamTitle':
                                        #print("Set meta")
                                        self.tauon.pctl.tag_meta = b.rstrip("'").lstrip("'")
                                        break
                        except:
                            r.close()
                            self.download_running = False
                            self.abort = True
                            self.tauon.gui.show_message("Data malformation detected. Stream aborted.", mode='error')
                            raise
        except Exception as e:
            print("Stream download thread crashed!")
            self.download_running = False
            self.abort = True
            return


import pychromecast
from t_modules.t_extra import shooter
import time
import socket

def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


class Chrome:
    def __init__(self, tauon):
        self.tauon = tauon
        self.services = []
        self.active = False
        self.cast = None
        self.target_playlist = None
        self.target_id = None
        self.save_vol = 100

    def rescan(self):
        print("Scanning for chromecasts...")

        if True: #not self.services:
            try:
                #self.tauon.gui.show_message(self.tauon.strings.scan_chrome)
                services, browser = pychromecast.discovery.discover_chromecasts()
                pychromecast.discovery.stop_discovery(browser)
                menu = self.tauon.chrome_menu
                MenuItem = self.tauon.MenuItem

                #menu.items.clear()
                for item in services:
                    self.services.append([str(item.uuid), str(item.friendly_name)])
                    menu.add_to_sub(1, MenuItem(self.tauon.strings.cast_to % str(item.friendly_name), self.three, pass_ref=True, args=[str(item.uuid), str(item.friendly_name)]))
                menu.add_to_sub(1, MenuItem(self.tauon.strings.stop_cast, self.end, show_test=lambda x: self.active))
            except:
                raise
                print("Failed to get chromecasts")


    def three(self, _, item):
        shooter(self.four, [item])

    def four(self, item):
        if self.active:
            self.end()
        self.tauon.start_remote()
        ccs, browser = pychromecast.get_listed_chromecasts(friendly_names=[item[1]], discovery_timeout=3.0)
        self.browser = browser
        self.cast = ccs[0]
        self.cast.wait()
        self.save_vol = self.tauon.pctl.player_volume
        self.tauon.pctl.player_volume = min(self.cast.status.volume_level * 100, 100)
        self.ip = get_ip()

        mc = self.cast.media_controller
        mc.app_id = "2F76715B"

        self.tauon.chrome_mode = True
        self.active = True
        self.tauon.gui.update += 1
        self.tauon.pctl.playerCommand = "startchrome"
        self.tauon.pctl.playerCommandReady = True
        self.tauon.tm.ready_playback()


    def update(self):
        self.cast.media_controller.update_status()
        return self.cast.media_controller.status.current_time, \
            self.cast.media_controller.status.media_custom_data.get("id"), \
            self.cast.media_controller.status.player_state, \
               self.cast.media_controller.status.duration

    def start(self, track_id, enqueue=False, t=0, url=None):
        self.cast.wait()
        tr = self.tauon.pctl.g(track_id)
        n = 0
        try:
            n = int(tr.track_number)
        except:
            pass
        d = {
            "metadataType": 3,
            "albumName": tr.album,
            "title": tr.title,
            "albumArtist": tr.album_artist,
            "artist": tr.artist,
            "trackNumber": n,
            "images": [{"url": f"http://{self.ip}:7814/api1/pic/medium/{track_id}"}],
            "releaseDate": tr.date
        }
        m = {
            "duration": round(float(tr.length), 1),
            "customData": {"id": str(tr.index)}
        }

        if url is None:
            url = f"http://{self.ip}:7814/api1/file/{track_id}"
        else:
            url = url.replace("localhost", self.ip)
            url = url.replace("127.0.0.1", self.ip)

        self.cast.media_controller.play_media(url, 'audio/mpeg', media_info=m, metadata=d, current_time=t, enqueue=enqueue)

    def stop(self):
        self.cast.media_controller.stop()

    def play(self):
        self.cast.media_controller.play()

    def pause(self):
        self.cast.media_controller.pause()

    def seek(self, t):
        self.cast.media_controller.seek(t)

    def volume(self, decimal):
        self.cast.set_volume(decimal)

    def end(self):
        self.tauon.pctl.playerCommand = "endchrome"
        self.tauon.pctl.playerCommandReady = True
        if self.active:
            if self.cast:
                mc = self.cast.media_controller
                mc.stop()
            self.active = False
        self.tauon.chrome_mode = False
        self.tauon.pctl.player_volume = self.save_vol




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import pickle
import sys
from ctypes import pointer
from gi.repository import GLib

if sys.platform != 'win32':
    import fcntl

n_version = "7.8.0"
t_version = "v" + n_version
t_title = 'Tauon'
t_id = 'tauonmb'
t_agent = "TauonMusicBox/" + n_version


print(f"{t_title} {t_version}")
print('Copyright 2015-2023 Taiko2k captain.gxj@gmail.com\n')

# Early arg processing
def transfer_args_and_exit():
    import urllib.request
    base = "http://localhost:7813/"

    if len(sys.argv) <= 1:
        url = base + "raise/"
        urllib.request.urlopen(url)

    for item in sys.argv:

        if not item.endswith(".py") and not item.startswith("-") and not item.endswith("exe") and (item.startswith("file://") or os.path.exists(item)):
            import base64
            url = base + "open/" + base64.urlsafe_b64encode(item.encode()).decode()
            urllib.request.urlopen(url)
        if item == "--play-pause":
            url = base + "playpause/"
            urllib.request.urlopen(url)
        if item == "--play":
            url = base + "play/"
            urllib.request.urlopen(url)
        if item == "--pause":
            url = base + "pause/"
            urllib.request.urlopen(url)
        if item == "--stop":
            url = base + "stop/"
            urllib.request.urlopen(url)
        if item == "--next":
            url = base + "next/"
            urllib.request.urlopen(url)
        if item == "--previous":
            url = base + "previous/"
            urllib.request.urlopen(url)
        if item == "--shuffle":
            url = base + "shuffle/"
            urllib.request.urlopen(url)
        if item == "--repeat":
            url = base + "repeat/"
            urllib.request.urlopen(url)

    sys.exit()


if "--no-start" in sys.argv:
    transfer_args_and_exit()

install_directory = os.path.dirname(os.path.abspath(__file__))

pyinstaller_mode = False
if hasattr(sys, "_MEIPASS"):
    pyinstaller_mode = True
if install_directory.endswith("\\_internal"):
    pyinstaller_mode = True
    install_directory = os.path.dirname(install_directory)

if pyinstaller_mode:
     os.environ["PATH"] += ":" + sys._MEIPASS
     os.environ['SSL_CERT_FILE'] = os.path.join(install_directory, "certifi", "cacert.pem")

# If we're installed, use home data locations
install_mode = False
if install_directory.startswith("/opt/")\
        or install_directory.startswith("/usr/")\
        or install_directory.startswith("/app/")\
        or install_directory.startswith("/snap/") or sys.platform == "darwin" or sys.platform == 'win32':
    install_mode = True

# Assume that it's a classic Linux install, use standard paths
if install_directory.startswith("/usr/"):
    install_directory = "/usr/share/TauonMusicBox"

user_directory = os.path.join(install_directory, "user-data")
config_directory = user_directory
asset_directory = os.path.join(install_directory, "assets")

if install_directory.startswith("/app/"):
    # Its Flatpak
    t_id = "com.github.taiko2k.tauonmb"
os.environ["SDL_VIDEO_WAYLAND_WMCLASS"] = t_id
os.environ["SDL_VIDEO_X11_WMCLASS"] = t_id

if os.path.isfile(os.path.join(install_directory, "portable")):
    install_mode = False

if install_mode:
    user_directory = os.path.join(GLib.get_user_data_dir(), "TauonMusicBox")
if not os.path.isdir(user_directory):
    os.makedirs(user_directory)

fp = None
dev_mode = os.path.isfile(os.path.join(install_directory, '.dev'))
if dev_mode:
    print("Dev mode, ignoring single instancing")
elif sys.platform != 'win32':
    pid_file = os.path.join(user_directory, 'program.pid')
    fp = open(pid_file, 'w')
    try:
        fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        # another instance is running
        print("Program is already running")
        transfer_args_and_exit()
else:
    if sys.platform == "win32":
        pid_file = os.path.join(user_directory, 'program.pid')
        try:
            if os.path.isfile(pid_file):
                os.remove(pid_file)
            fp = open(pid_file, 'w')
        except IOError:
            # another instance is running
            print("Program is already running")
            transfer_args_and_exit()
    if pyinstaller_mode:
        os.environ["FONTCONFIG_PATH"] = os.path.join(install_directory, "etc\\fonts")#"C:\\msys64\\mingw64\\etc\\fonts"

phone = False
d = os.environ.get('XDG_CURRENT_DESKTOP')
if d in ["GNOME:Phosh"]:
    os.environ["SDL_VIDEODRIVER"] = "wayland"
    phone = True

if pyinstaller_mode: # and sys.platform == 'darwin':
    os.environ["PYSDL2_DLL_PATH"] = install_directory

fs_mode = False
if os.environ.get('GAMESCOPE_WAYLAND_DISPLAY') is not None:
    fs_mode = True
    print("Running in GAMESCOPE MODE")

allow_hidpi = True
if sys.platform == "win32" and sys.getwindowsversion().major < 10 or os.path.isfile(os.path.join(user_directory, "nohidpi")):
    allow_hidpi = False

from sdl2 import *
from sdl2.sdlimage import *

SDL_SetHint(SDL_HINT_VIDEO_ALLOW_SCREENSAVER, b'1')
SDL_SetHint(SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH, b"1")
SDL_SetHint(SDL_HINT_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR, b"0")
SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "best".encode())

draw_border = True
w = 1120
h = 600
if phone:
    w = 720
    h = 1800
window_default_size = [w, h]
window_size = [w, h]
logical_size = [w, h]
window_opacity = 1
scale = 1
if sys.platform == "darwin":
    scale = 2
if phone:
    scale = 1.3

maximized = False
old_window_position = None

window_p = os.path.join(user_directory, "window.p")
if os.path.isfile(window_p) and not fs_mode:
    try:
        state_file = open(window_p, "rb")
        save = pickle.load(state_file)
        state_file.close()

        draw_border = save[0]
        window_size = save[1]
        w, h = save[1]
        if 100 < w < 10000 and 100 < h < 5000:
            logical_size[0], logical_size[1] = w, h
        window_opacity = save[2]
        scale = save[3]
        maximized = save[4]
        old_window_position = save[5]
        del save

    except:
        print('Corrupted window state file?!')
        print('Please restart app')
        os.remove(window_p)
        exit(1)
else:
    print("No window state file")


if d == "GNOME": #and os.environ.get("XDG_SESSION_TYPE") and os.environ.get("XDG_SESSION_TYPE") == "wayland":
    try:
        import gi.repository
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk

        gtk_settings = Gtk.Settings().get_default()
        xtheme = gtk_settings.get_property("gtk-cursor-theme-name")
        xsize = gtk_settings.get_property("gtk-cursor-theme-size")
        os.environ["XCURSOR_THEME"] = xtheme
        os.environ["XCURSOR_SIZE"] = str(xsize)
    except:
        pass

if os.environ.get("XDG_SESSION_TYPE") and os.environ.get("XDG_SESSION_TYPE") == "wayland":
    os.environ["SDL_VIDEODRIVER"] = "wayland"
if os.path.exists(os.path.join(user_directory, "x11")):
    os.environ["SDL_VIDEODRIVER"] = "x11"

SDL_Init(SDL_INIT_VIDEO)

err = SDL_GetError()
if err and "GLX" in err.decode():
    print(f"SDL init error: {err.decode()}")
    SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, b"Tauon Music Box failed to start :(",
                             b"Error: " + err + b".\n If you're using Flatpak, try run `$ flatpak update`", None)
    sys.exit()

window_title = t_title
window_title = window_title.encode('utf-8')

flags = SDL_WINDOW_RESIZABLE

if allow_hidpi:
    flags |= SDL_WINDOW_ALLOW_HIGHDPI

if draw_border and not fs_mode:
    flags |= SDL_WINDOW_BORDERLESS

if fs_mode:
    flags |= SDL_WINDOW_FULLSCREEN_DESKTOP

if old_window_position is None:
    o_x = SDL_WINDOWPOS_UNDEFINED
    o_y = SDL_WINDOWPOS_UNDEFINED
else:
    o_x = old_window_position[0]
    o_y = old_window_position[1]

if "--tray" in sys.argv:
    flags |= SDL_WINDOW_HIDDEN

error = False

t_window = SDL_CreateWindow(window_title,
                            o_x, o_y,
                            logical_size[0], logical_size[1],
                            flags) # | SDL_WINDOW_FULLSCREEN)



if maximized:
    SDL_MaximizeWindow(t_window)


renderer = SDL_CreateRenderer(t_window, 0, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC)

if not renderer or not t_window:
    print("ERROR CREATING WINDOW!")

SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_BLEND)
SDL_SetWindowOpacity(t_window, window_opacity)

SDL_SetRenderDrawColor(renderer, 7, 7, 7, 255)
SDL_RenderClear(renderer)


i_x = pointer(c_int(0))
i_y = pointer(c_int(0))
SDL_GL_GetDrawableSize(t_window, i_x, i_y)
window_size[0] = i_x.contents.value
window_size[1] = i_y.contents.value

SDL_GetWindowSize(t_window, i_x, i_y)
logical_size[0] = i_x.contents.value
logical_size[1] = i_y.contents.value

img_path = os.path.join(asset_directory, "loading.png")
if scale != 1:
    img_path2 = os.path.join(user_directory, "scaled-icons", "loading.png")
    if os.path.isfile(img_path2):
        img_path = img_path2
    del img_path2

raw_image = IMG_Load(img_path.encode())
sdl_texture = SDL_CreateTextureFromSurface(renderer, raw_image)
w = raw_image.contents.w
h = raw_image.contents.h
rect = SDL_Rect(window_size[0] // 2 - w // 2, window_size[1] // 2 - h // 2, w, h)

SDL_RenderCopy(renderer, sdl_texture, None, rect)

SDL_RenderPresent(renderer)

SDL_FreeSurface(raw_image)
SDL_DestroyTexture(sdl_texture)

from t_modules import t_bootstrap
h = t_bootstrap.holder
h.w = t_window
h.r = renderer
h.wl = logical_size
h.wr = window_size
h.wdf = window_default_size
h.s = scale
h.m = maximized
h.e = transfer_args_and_exit
h.d = draw_border
h.o = window_opacity
h.ow = old_window_position
h.id = install_directory
h.py = pyinstaller_mode
h.p = phone
h.window_title = window_title
h.fs_mode = fs_mode
h.title = t_title
h.n_version = n_version
h.t_version = t_version
h.t_id = t_id
h.agent = t_agent
h.dev_mode = dev_mode
h.lock = fp

del raw_image
del sdl_texture
del w
del h
del rect
del flags
del img_path

if pyinstaller_mode or sys.platform == "darwin" or install_mode:
    from t_modules import t_main
else:
    # Using the above import method breaks previous pickles. Could be fixed
    # but yet to decide what best method is.
    big_boy_path = os.path.join(install_directory, 't_modules/t_main.py')
    f = open(big_boy_path, "rb")
    main = compile(f.read(), big_boy_path, 'exec')
    f.close()
    del big_boy_path
    del f
    exec(main)





# Tauon Music Box - Misc Functions Module

# Copyright © 2015-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


import time
import random
import colorsys
import subprocess
import os
import shlex
import zipfile
import glob
import locale
import re
import math
import threading
import urllib.parse
import gi
from gi.repository import GLib

def tmp_cache_dir():
    tmp_dir = GLib.get_tmp_dir()
    return os.path.join(tmp_dir, "TauonMusicBox")

# A seconds based timer
class Timer:
    def __init__(self, force=None):
        self.start = 0
        self.end = 0
        self.set()
        if force:
            self.force_set(force)

    def set(self):  # Reset
        self.start = time.monotonic()

    def hit(self):  # Return time and reset

        self.end = time.monotonic()
        elapsed = self.end - self.start
        self.start = time.monotonic()
        return elapsed

    def get(self):  # Return time only
        self.end = time.monotonic()
        return self.end - self.start

    def force_set(self, sec):
        self.start = time.monotonic()
        self.start -= sec


# Simple bool timer object
class TestTimer:
    def __init__(self, time):
        self.timer = Timer()
        self.time = time

    def test(self):
        return self.timer.get() > self.time

j_chars = "あおいえうんわらまやはなたさかみりひにちしきるゆむぬつすくれめへねてせけをろもほのとそこアイウエオンヲラマハナタサカミヒニチシキルユムフヌツスクレメヘネテセケロヨモホノトソコ"


# Test given proximity between two 2d points to given square
def point_proximity_test(a, b, p):
    return abs(a[0] - b[0]) < p and abs(a[1] - b[1]) < p

# Get distance between two points
def point_distance(a, b):
    return math.sqrt(abs(a[0] - b[0]) ** 2 + abs(b[1] - b[1]) ** 2)

# Removes whatever this is from a line, I forgot
def rm_16(line):
    if "ÿ þ" in line:
        for c in line:
            line = line[1:]
            if c == 'þ':
                break

        line = line[::2]
    return line


# Returns a string from seconds to a compact time format, e.g 2h:23
def get_display_time(seconds):
    result = divmod(int(seconds), 60)
    if result[0] > 99:
        result = divmod(result[0], 60)
        return str(result[0]) + 'h ' + str(result[1]).zfill(2)
    return str(result[0]).zfill(2) + ":" + str(result[1]).zfill(2)


def get_hms_time(seconds):
    m, s = divmod(round(seconds), 60)
    h, m = divmod(m, 60)
    if h:
        return f'{h:d}:{m:02d}:{s:02d}'
    else:
        return f'{m:02d}:{s:02d}'

def hms_to_seconds(time_str):
    components = time_str.split(':')
    seconds = 0
    if len(components) == 1:  # If only seconds provided
        seconds = int(components[0])
    elif len(components) == 2:  # If minutes and seconds provided
        seconds = int(components[0]) * 60 + int(components[1])
    elif len(components) == 3:  # If hours, minutes, and seconds provided
        seconds = int(components[0]) * 3600 + int(components[1]) * 60 + int(components[2])
    return seconds

# Creates a string from number of bytes to X MB/kB etc with Locale adjustment
def get_filesize_string(file_bytes, rounding=2):
    if not file_bytes:
        return "0"
    if file_bytes < 1000:
        line = locale.str(file_bytes) + " B"
    elif file_bytes < 1000000:
        file_kb = round(file_bytes / 1000, rounding)
        line = locale.str(file_kb) + " KB"
    elif file_bytes < 1000000000:
        file_mb = round(file_bytes / 1000000, rounding)
        line = locale.str(file_mb) + " MB"
    else:
        file_mb = round(file_bytes / 1000000000, 1)
        line = locale.str(file_mb) + " GB"
    return line

def get_filesize_string_rounded(file_bytes):
    if not file_bytes:
        return "0"
    if file_bytes < 1000:
        line = str(round(file_bytes)) + " B"
    elif file_bytes < 1000000:
        file_kb = round(file_bytes / 1000)
        line = str(file_kb) + " KB"
    else:
        file_mb = round(file_bytes / 1000000, 1)
        line = str(file_mb) + " MB"
    return line

# Estimates the perceived luminance of a colour
def test_lumi(c1):
    return 1 - (0.299 * c1[0] + 0.587 * c1[1] + 0.114 * c1[2]) / 255

def rel_luminance(colour):
    r = colour[0] / 255
    g = colour[1] / 255
    b = colour[2] / 255

    if r < 0.03928:
        r /= 12.90
    else:
        r = ((r + 0.055) / 1.055) ** 2.4

    if g < 0.03928:
        g /= 12.90
    else:
        g = ((g + 0.055) / 1.055) ** 2.4

    if b < 0.03928:
        b /= 12.90
    else:
        b = ((b + 0.055) / 1.055) ** 2.4

    return 0.2126 * r + 0.7152 * g + 0.0722 * b


def contrast_ratio(c1, c2):

    l1 = rel_luminance(c1)
    l2 = rel_luminance(c2)

    if l2 > l1:
        l2, l1 = l1, l2

    return (l1 + 0.05) / (l2 + 0.05)

# Gives the sum of first 3 elements in a list
def colour_value(c1):
    return c1[0] + c1[1] + c1[2]


# Performs alpha blending of one colour (rgba) onto another (rgb)
def alpha_blend(colour, base):
    alpha = colour[3] / 255
    return [int(alpha * colour[0] + (1 - alpha) * base[0]),
            int(alpha * colour[1] + (1 - alpha) * base[1]),
            int(alpha * colour[2] + (1 - alpha) * base[2]), 255]


# Change the alpha component of an RGBA list
def alpha_mod(colour, alpha):
    return [colour[0], colour[1], colour[2], alpha]


# Shift between two colours based on x where x is between 0 and limit
def colour_slide(a, b, x, x_limit):

    return (min(int(a[0] + ((b[0] - a[0]) * (x / x_limit))), 255),
     min(int(a[1] + ((b[1] - a[1]) * (x / x_limit))), 255),
     min(int(a[2] + ((b[2] - a[2]) * (x / x_limit))), 255), 255)


def hex_to_rgb(colour):
    colour = colour.strip("#")
    return list(int(colour[i:i + 2], 16) for i in (0, 2, 4)) + [255]


# Checks if all the numbers in a list are the same
def check_equal(lst):
    return not lst or lst.count(lst[0]) == len(lst)


# Check if the first 3 elements of a list are the same
def is_grey(lst):
    return lst[0] == lst[1] == lst[2]


# Gives a score from 0-7 based on number of seconds
def star_count(sec, dur):
    stars = 0
    if dur and sec / dur > 0.95:
        stars += 1
    if sec > 60 * 15:
        stars += 1
    if sec > 60 * 30:
        stars += 1
    if sec > 60 * 60:
        stars += 1
    if sec > 60 * 60 * 2:
        stars += 1
    if sec > 60 * 60 * 4:
        stars += 1
    if sec > 60 * 60 * 8:
        stars += 1
    if sec > 60 * 60 * 16:
        stars += 1
    return stars


def star_count3(sec, dur):
    stars = 0
    if dur and sec / dur > 0.95:
        stars += 1
    if dur and sec / dur > 1.95:
        stars += 1
    if sec > 60 * 5:
        stars += 1
    if sec > 60 * 10:
        stars += 1
    if sec > 60 * 15:
        stars += 1
    if sec > 60 * 20:
        stars += 1
    if sec > 60 * 30:
        stars += 1
    if sec > 60 * 60:
        stars += 1
    if sec > 60 * 60 * 1.5:
        stars += 1
    if sec > 60 * 60 * 1.75:
        stars += 1
    # if sec > 60 * 60 * 2:
    #     stars += 1
    return stars


# Gives a score from 0.0 - 1.0 based on number of seconds
def star_count2(sec):
    star = 0
    star += min(sec / (60 * 4), 0.2)
    star += min(sec / (60 * 60), 0.4)
    star += sec / (60 * 60 * 10)
    return float(round(min(star, 1), 1))


def search_magic(terms, evaluate):
    return all(word in evaluate for word in terms.split())


def search_magic_any(terms, evaluate):
    return any(word in evaluate for word in terms.split())


def random_colour(saturation, luminance):

    h = round(random.random(), 2)
    colour = colorsys.hls_to_rgb(h, luminance, saturation)
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255]


def hsl_to_rgb(h, s, l):
    colour = colorsys.hls_to_rgb(h, l, s)
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255]

def hls_to_rgb(h, l, s):
    colour = colorsys.hls_to_rgb(h, l, s)
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), 255]

def rgb_to_hls(r, g, b):
    return colorsys.rgb_to_hls(r / 255, g / 255, b / 255)

def rgb_add_hls(source, h=0, l=0, s=0):
    c = colorsys.rgb_to_hls(source[0] / 255, source[1] / 255, source[2] / 255)
    colour = colorsys.hls_to_rgb(c[0] + h, min(max(c[1] + l, 0), 1), min(max(c[2] + s, 0), 1))
    return [int(colour[0] * 255), int(colour[1] * 255), int(colour[2] * 255), source[3]]


def is_light(colour):
    return test_lumi(colour) < 0.2

class ColourGenCache:

    def __init__(self, saturation, luminance):

        self.saturation = saturation
        self.luminance = luminance
        self.store = {}

    def get(self, key):

        if key in self.store:
            return self.store[key]

        colour = random_colour(self.saturation, self.luminance)

        self.store[key] = colour
        return colour




def folder_file_scan(path, extensions):

    match = 0
    count = sum([len(files) for r, d, files in os.walk(path)])
    for ext in extensions:

        match += len(glob.glob(path + '/**/*.' + ext.lower(), recursive=True))

    if count == 0:
        return 0

    if count < 5 and match > 0:
        return 1

    return match / count


def is_ignorable_file(string):
    for s in [
        "Thumbs.db",
        ".log",
        "desktop.ini",
        "DS_Store",
        ".nfo",
        "yric"
    ]:
        if s in string:
            return True
    return False


# Pre-compile the regular expression pattern for dates starting with the year
date_pattern = re.compile(r'\b(?:\d{2}([/. -])\d{2}\1(\d{4})|\b(\d{4})([/. -])\d{2}\4\d{2}).*')

def get_year_from_string(s):
    # Search for the pattern in the string
    match = date_pattern.search(s)

    # Extract and return the year if a match is found
    if match:
        return match.group(2) if match.group(2) else match.group(3)

    return ""


# Example usage
example_string = "Event date: 2021-12-31."
print(get_year_from_string(example_string))  # Output: "2021"

def is_music_related(string):
    for s in [
        "Folder.jpg",
        "folder.jpg",
        "Cover.jpg",
        "cover.jpg",
        "AlbumArt",
        ".m3u",
        ".m3u8",
        ".cue",
        ".CUE",
    ]:
        if s in string:
            return True
    return False


# Get ratio of given file extensions in archive
def archive_file_scan(path, extensions, launch_prefix=""):

    ext = os.path.splitext(path)[1][1:].lower()
    # print(path)
    # print(ext)
    try:
        if ext == 'rar':
            matches = 0
            count = 0
            line = launch_prefix + "unrar lb -p- " + shlex.quote(path) + " " + shlex.quote(os.path.dirname(path)) + os.sep
            result = subprocess.run(shlex.split(line), stdout=subprocess.PIPE)
            file_list = result.stdout.decode("utf-8", 'ignore').split("\n")
            # print(file_list)
            for fi in file_list:
                for ty in extensions:
                    if fi[len(ty) * -1:].lower() == ty:
                        matches += 1
                        break
                    elif is_ignorable_file(fi):
                        count -= 1
                        break
                    elif is_music_related(fi):
                        matches += 5
                count += 1
            if count > 200:
                #print("RAR archive has many files")
                #print("   --- " + path)
                return 0
            if matches == 0:
                #print("RAR archive does not appear to contain audio files")
                #print("   --- " + path)
                return 0
            if count == 0:
                #print("Archive has no files")
                #print("   --- " + path)
                return 0

        elif ext == '7z':
            matches = 0
            count = 0
            line = launch_prefix + "7z l " + shlex.quote(path) # + " " + shlex.quote(os.path.dirname(path)) + os.sep
            result = subprocess.run(shlex.split(line), stdout=subprocess.PIPE)
            file_list = result.stdout.decode("utf-8", 'ignore').split("\n")
            # print(file_list)

            for fi in file_list:

                if '....A' not in fi:
                    continue
                for ty in extensions:
                    if fi[len(ty) * -1:].lower() == ty:
                        matches += 1
                        break
                    elif is_ignorable_file(fi):
                        count -= 1
                        break
                    elif is_music_related(fi):
                        matches += 5
                count += 1

            if count > 200:
                #print("7z archive has many files")
                #print("   --- " + path)
                return 0
            if matches == 0:
                #print("7z archive does not appear to contain audio files")
                #print("   --- " + path)
                return 0
            if count == 0:
                #print("7z archive has no files")
                #print("   --- " + path)
                return 0

        elif ext == "zip":

            zip_ref = zipfile.ZipFile(path, 'r')
            matches = 0
            count = 0
            #print(zip_ref.namelist())
            for fi in zip_ref.namelist():
                for ty in extensions:
                    if fi[len(ty) * -1:].lower() == ty:
                        matches += 1
                        break
                    elif is_ignorable_file(fi):
                        count -= 1
                        break
                    elif is_music_related(fi):
                        matches += 5
                count += 1
            if count == 0:
                #print("Archive has no files")
                #print("   --- " + path)
                return 0
            if count > 300:
                #print("Zip archive has many files")
                #print("   --- " + path)
                return 0
            if matches == 0:
                #print("Zip archive does not appear to contain audio files")
                #print("   --- " + path)
                return 0
        else:
            return 0

    except:
        print("Archive test error")

        return 0

    if count == 0:
        return 0

    ratio = matches / count
    if count < 5 and matches > 0:
        ratio = 100
    return ratio


def get_folder_size(path):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size


def filename_safe(text, sub=""):
    for cha in '/\\<>:"|?*':
        text = text.replace(cha, sub)
    return text.rstrip(" .")

def filename_to_metadata(filename):

    # Remove the file extension
    name_without_extension = filename.rsplit('.', 1)[0]

    # Remove leading track numbers if present
    name_without_track_number = re.sub(r'^\d{1,2}[. -]+', '', name_without_extension)

    # Split the filename on ' - ' to separate the artist and the title
    if ' - ' in name_without_track_number:
        artist, title = name_without_track_number.split(' - ', 1)
    else:
        # If ' - ' is not present, return the whole name as title and empty string as artist
        artist = ''
        title = name_without_track_number

    return artist, title


def get_artist_strip_feat(track_object):

    artist_name = track_object.artist #.lower()
    if track_object.album_artist:
        if "feat." in artist_name or "pres." in artist_name or ", " in artist_name or "; " in artist_name or not artist_name:
            if track_object.album_artist.lower() != "va" and \
                    track_object.album_artist.lower() != "various artists":
                artist_name = track_object.album_artist
    return artist_name

def get_artist_safe(track):

    if track:
        artist = track.album_artist
        if not artist:
            artist = track.artist
        artist = filename_safe(artist)
        artist = artist.split("feat")[0]
        artist = artist.split(", ")[0]
        artist = artist.split("; ")[0]
        return artist
    return ""

def get_split_artists(track):
    if 'artists' in track.misc:
        return track.misc['artists']
    artist = track.artist.split("feat")[0].strip()
    return re.split(r'; |, |& ', artist)

def coll_rect(rect1, rect2):

    if rect1[0] + rect1[2] < rect2[0] or \
            rect1[1] + rect1[3] < rect2[1] or \
            rect1[0] > rect2[0] + rect2[2] or \
            rect1[1] > rect2[1] + rect2[3]:
        return False
    return True


def commonprefix(l):

    cp = []
    ls = [p.split('/') for p in l]
    ml = min(len(p) for p in ls)

    for i in range(ml):

        s = set(p[i] for p in ls)
        if len(s) != 1:
            break

        cp.append(s.pop())

    return '/'.join(cp)


def fader_timer(time_point, start, duration, off=True, range=255):

    if time_point < start:
        fade = range
    elif time_point < start + duration:
        p = (time_point - start) / duration
        fade = int(range - (range * p))
    else:
        fade = 0

    return fade


id3_genre_dict = {
    
    0: "Blues",
    1: "Classic Rock",
    2: "Country",
    3: "Dance",
    4: "Disco",
    5: "Funk",
    6: "Grunge",
    7: "Hip-Hop",
    8: "Jazz",
    9: "Metal",
    10: "New Age",
    11: "Oldies",
    12: "Other",
    13: "Pop",
    14: "R&B",
    15: "Rap",
    16: "Reggae",
    17: "Rock",
    18: "Techno",
    19: "Industrial",
    20: "Alternative",
    21: "Ska",
    22: "Death Metal",
    23: "Pranks",
    24: "Soundtrack",
    25: "Euro-Techno",
    26: "Ambient",
    27: "Trip-Hop",
    28: "Vocal",
    29: "Jazz+Funk",
    30: "Fusion",
    31: "Trance",
    32: "Classical",
    33: "Instrumental",
    34: "Acid",
    35: "House",
    36: "Game",
    37: "Sound Clip",
    38: "Gospel",
    39: "Noise",
    40: "Alternative Rock",
    41: "Bass",
    42: "Soul",
    43: "Punk",
    44: "Space",
    45: "Meditative",
    46: "Instrumental Pop",
    47: "Instrumental Rock",
    48: "Ethnic",
    49: "Gothic",
    50: "Darkwave",
    51: "Techno-Industrial",
    52: "Electronic",
    53: "Pop-Folk",
    54: "Eurodance",
    55: "Dream",
    56: "Southern Rock",
    57: "Comedy",
    58: "Cult",
    59: "Gangsta Rap",
    60: "Top 40",
    61: "Christian Rap",
    62: "Pop/Funk",
    63: "Jungle",
    64: "Native American",
    65: "Cabaret",
    66: "New Wave",
    67: "Psychedelic",
    68: "Rave",
    69: "Showtunes",
    70: "Trailer",
    71: "Lo-Fi",
    72: "Tribal",
    73: "Acid Punk",
    74: "Acid Jazz",
    75: "Polka",
    76: "Retro",
    77: "Musical",
    78: "Rock & Roll",
    79: "Hard Rock",
    80: "Folk",
    81: "Folk-Rock",
    82: "National Folk",
    83: "Swing",
    84: "Fast Fusion",
    85: "Bebob",
    86: "Latin",
    87: "Revival",
    88: "Celtic",
    89: "Bluegrass",
    90: "Avantgarde",
    91: "Gothic Rock",
    92: "Progressive Rock",
    93: "Psychedelic Rock",
    94: "Symphonic Rock",
    95: "Slow Rock",
    96: "Big Band",
    97: "Chorus",
    98: "Easy Listening",
    99: "Acoustic",
    100: "Humour",
    101: "Speech",
    102: "Chanson",
    103: "Opera",
    104: "Chamber Music",
    105: "Sonata",
    106: "Symphony",
    107: "Booty Bass",
    108: "Primus",
    109: "Porn Groove",
    110: "Satire",
    111: "Slow Jam",
    112: "Club",
    113: "Tango",
    114: "Samba",
    115: "Folklore",
    116: "Ballad",
    117: "Power Ballad",
    118: "Rhythmic Soul",
    119: "Freestyle",
    120: "Duet",
    121: "Punk Rock",
    122: "Drum Solo",
    123: "A Cappella",
    124: "Euro-House",
    125: "Dance Hall",
    126: "Goa",
    127: "Drum & Bass",
    128: "Club-House",
    129: "Hardcore",
    130: "Terror",
    131: "Indie",
    132: "BritPop",
    133: "Negerpunk",
    134: "Polsk Punk",
    135: "Beat",
    136: "Christian Gangsta Rap",
    137: "Heavy Metal",
    138: "Black Metal",
    139: "Crossover",
    140: "Contemporary Christian",
    141: "Christian Rock",
    142: "Merengue",
    143: "Salsa",
    144: "Thrash Metal",
    145: "Anime",
    146: "JPop",
    147: "Synthpop",
    148: "Abstract",
    149: "Art Rock",
    150: "Baroque",
    151: "Bhangra",
    152: "Big Beat",
    153: "Breakbeat",
    154: "Chillout",
    155: "Downtempo",
    156: "Dub",
    157: "EBM",
    158: "Eclectic",
    159: "Electro",
    160: "Electroclash",
    161: "Emo",
    162: "Experimental",
    163: "Garage",
    164: "Global",
    165: "IDM",
    166: "Illbient",
    167: "Industro-Goth",
    168: "Jam Band",
    169: "Krautrock",
    170: "Leftfield",
    171: "Lounge",
    172: "Math Rock",
    173: "New Romantic",
    174: "Nu-Breakz",
    175: "Post-Punk",
    176: "Post-Rock",
    177: "Psytrance",
    178: "Shoegaze",
    179: "Space Rock",
    180: "Trop Rock",
    181: "World Music",
    182: "Neoclassical",
    183: "Audiobook",
    184: "Audio Theatre",
    185: "Neue Deutsche Welle",
    186: "Podcast",
    187: "Indie Rock",
    188: "G-Funk",
    189: "Dubstep",
    190: "Garage Rock",
    191: "Psybient",
    192: "Unknown",
}


class FunctionStore:
    # Stores functions and arguments for calling later

    def __init__(self):

        self.items = []

    def store(self, function, args=()):

        self.items.append((function, args))

    def recall_all(self):

        while self.items:
            item = self.items.pop()
            item[0](*item[1])



def grow_rect(rect, px):
    return rect[0] - px, rect[1] - px, rect[2] + px * 2, rect[3] + px * 2

# def get_hash(f_path, mode='sha256'):
#     h = hashlib.new(mode)
#     with open(f_path, 'rb') as file:
#         data = file.read()
#     h.update(data)
#     digest = h.hexdigest()
#     return digest

def subtract_rect(base, hole):  # Return 4 rects from 1 minus 1 inner (with overlaps)

    west = base[0], base[1], hole[0], base[3]
    north = base[0], base[1], base[2], hole[1] - base[1]
    east = base[0] + hole[0] + hole[2], base[1], base[2] - (hole[0] + hole[2]), base[3]
    south = base[0], hole[1] + hole[3], base[2], base[3] - hole[3] - 2

    return west, north, east, south

genre_corrections = [
    "J-Pop",
    "J-Rock",
    "K-Pop",
    "Hip Hop",
]

genre_corrections2 = [x.lower().replace("-", "").replace(" ", "") for x in genre_corrections]

def genre_correct(text):

    parsed = text.lower().replace("-", "").replace(" ", "").strip()
    if parsed.startswith("post"):
        return ("Post-" + parsed[4:]).title()
    if parsed in genre_corrections2:
        return genre_corrections[genre_corrections2.index(parsed)]
    return text.title().strip()


def reduce_paths(paths):  # in-place remove of redundant sub-paths from list of folder paths

    paths[:] = list(set(paths))[:]  # remove duplicates

    while "" in paths:
        paths.remove("")

    while True:
        remove_path = False
        for i in reversed(range(len(paths))):
            path = paths[i].rstrip("\\/")

            for b in reversed(range(len(paths))):
                path2 = paths[b].rstrip("\\/")

                if len(path) > len(path2) and path.startswith(path2) and path[len(path2)] in ("/", "\\"):
                    del paths[i]
                    remove_path = True
                    break

            if remove_path:
                break
        if not remove_path:
            break

def fit_box(inner, outer):
    scale = min(outer[0]/inner[0], outer[1]/inner[1])
    return round(inner[0] * scale), round(inner[1] * scale)


def seconds_to_day_hms(seconds, s_day, s_days):

    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minuites, seconds = divmod(seconds, 60)

    if days == 1:
        return f"{str(int(days))} {s_day}, {str(int(hours))}:{str(int(minuites))}:{str(int(seconds))}"
    else:
        return f"{str(int(days))} {s_days}, {str(int(hours))}:{str(int(minuites)).zfill(2)}:{str(int(seconds)).zfill(2)}"


def shooter(func, args=()):
    shoot = threading.Thread(target=func, args=args)
    shoot.daemon = True
    shoot.start()


year_search = re.compile(r'\d{4}')

def d_date_display(track):
    if 'rdat' in track.misc:
        return str(track.date) + " → " + track.misc["rdat"]
    return str(track.date)

def d_date_display2(track):
    if 'rdat' in track.misc:
        return str(get_year_from_string(track.date)) + " → " + get_year_from_string(track.misc["rdat"])
    return str(get_year_from_string(track.date))

def process_odat(nt, odat):
    if odat and odat != nt.date and odat != nt.date[:4] and odat != nt.date[-4:] \
            and nt.date != odat[:4] and nt.date != odat[-4:]:
        if not nt.date:
            nt.date = odat
        else:
            nt.misc["rdat"] = nt.date
            nt.date = odat

def clean_string(s):
    return s.encode('utf-8', 'surrogatepass').decode('utf-8', 'replace')

def uri_parse(s):
    if s.startswith("file://"):
        s = str(urllib.parse.unquote(s)).replace("file://", "").replace("\r", "")
    return s

mac_styles = {
    "mac": None,
    "whitesur": None,
    "vimix": None,
    "sweet": None,
    "dracula": [[248, 58, 67, 255], [239, 251, 122, 255], [74, 254, 104, 255]],
    "nordic": None,
    "juno": None
}

def sleep_timeout(condition_function, time_limit=2):
    if condition_function():
        return
    timer = Timer()
    timer.set()
    while condition_function():
        time.sleep(0.01)
        if timer.get() > time_limit:
            break



from setuptools import setup, find_packages, Extension

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name = 'Tauon Music Box',
    packages = find_packages(),
    version = '7.7.2',
    url = 'https://tauonmusicbox.rocks/',
    license = 'GPL-3',
    install_requires=required,
	classifiers=[
		'Programming Language :: Python :: 3'
	],
)



# Tauon Music Box - Phazor audio backend module

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

import ctypes
from ctypes import *
from ctypes.util import find_library
import os.path
import time
import requests
from requests.models import PreparedRequest
import threading
import shutil
from t_modules.t_extra import *
import mutagen
import hashlib

def get_phazor_pathname(pctl):
    n = find_library("phazor")
    if n:
        return n

    n = os.path.join(pctl.install_directory, "lib/libphazor.so")
    if os.path.isfile(n):
        return n

def phazor_exists(pctl):
    if get_phazor_pathname(pctl):
        return True

    return False

def player4(tauon):

    pctl = tauon.pctl
    gui = tauon.gui
    prefs = tauon.prefs

    print("Start PHAzOR backend...")

    # # Get output device names
    # if len(prefs.phazor_devices) < 2 and not tauon.macos and not tauon.msys:
    #     try:
    #         import pulsectl
    #         pulse = pulsectl.Pulse('Tauon Music Box')
    #         sink_list = pulse.sink_list()
    #         for sink in sink_list:
    #             prefs.phazor_devices[sink.description] = sink.name
    #         pulse.close()
    #     except:
    #         print("Warning: Missing dependency Pulsectl")

    state = 0
    player_timer = Timer()
    loaded_track = None
    fade_time = 400

    aud = ctypes.cdll.LoadLibrary(get_phazor_pathname(pctl))

    aud.init()

    aud.get_device.restype = ctypes.c_char_p

    aud.feed_raw.argtypes = (ctypes.c_int,ctypes.c_char_p)
    aud.feed_raw.restype = None
    tauon.aud = aud
    aud.set_volume(int(pctl.player_volume))

    bins1 = (ctypes.c_float * 24)()
    bins2 = (ctypes.c_float * 45)()

    aud.get_level_peak_l.restype = ctypes.c_float
    aud.get_level_peak_r.restype = ctypes.c_float

    active_timer = Timer()

    def scan_device():
        n = aud.scan_devices()
        devices = ["Default"]
        if n:
            for d in range(n):
                st = aud.get_device(d).decode()
                devices.append(st)
        prefs.phazor_devices = devices
        if prefs.phazor_device_selected not in devices:
            prefs.phazor_device_selected = devices[0]

    scan_device()

    class LibreSpot:
        def __init__(self):

            self.running = False
            self.flush = False
        def go(self, force=False):
            aud.config_set_feed_samplerate(44100)
            aud.config_set_min_buffer(1000)
            if not shutil.which("librespot"):
                gui.show_message("SPP: Error, librespot not found")
                return 1
            # if not prefs.spot_username or not prefs.spot_password:
            #     gui.show_message("Please enter your spotify username and password in settings")
            #     return 1
            if tauon.librespot_p:
                if force:
                    print("SPP: Force restart librespot")
                    self.end()
                    tauon.librespot_p = None
                else:
                    self.flush = True

            pctl.spot_playing = True

            if not tauon.librespot_p:
                print("SPP: Librespot not running")
                username = tauon.spot_ctl.get_username()
                if not username or not prefs.spotify_token:
                    print("SPP: Error. Missing auth data")
                    return 1

                cache = os.path.join(tauon.cache_directory, "lsspot")
                if not os.path.exists(cache):
                    os.makedirs(cache)

                cmd = ["librespot", "--username", username, "--password", prefs.spot_password, "--backend", "pipe", "-n", "Tauon Music Box", "--disable-audio-cache", "--device-type", "computer", "--volume-ctrl", "fixed", "--initial-volume", "100"]
                #tauon.spot_ctl.preparing_spotify = True
                gui.update += 1

                startupinfo = None
                if tauon.msys:
                    startupinfo = subprocess.STARTUPINFO()
                    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                try:
                    tauon.librespot_p = subprocess.Popen(cmd, stdout=subprocess.PIPE, startupinfo=startupinfo)
                    print("SPP: Librespot now started")
                except Exception as e:
                    print(str(e))
                    print("SPP: Error starting librespot")
                    #tauon.spot_ctl.preparing_spotify = False
                    gui.update += 1
                    return 1

            return 0
        def soft_end(self):
            self.running = False
            pctl.spot_playing = False
        def end(self):
            self.running = False
            pctl.spot_playing = False
            if tauon.librespot_p:
                if tauon.msys:
                    tauon.librespot_p.terminate()
                    #tauon.librespot_p.communicate()
                else:
                    import signal
                    tauon.librespot_p.send_signal(signal.SIGINT)  # terminate() doesn't work

        def worker(self):
            self.running = True
            print("SPP: Enter librespot feeder thread")
            while True:

                if self.running is False:
                    print("SPP: Exit Librespot worker thread")
                    return
                if self.flush:
                    self.flush = False
                    print("SPP: Flushing some data...")
                    tauon.librespot_p.stdout.read(70000)  # rough
                    print("SPP: Done flush")
                if aud.feed_ready(1000) == 1:
                    if aud.get_buff_fill() < 2000:
                        data = tauon.librespot_p.stdout.read(1000)
                        aud.feed_raw(len(data), data)

                if state == 0:
                    tauon.librespot_p.stdout.read(50)
                    time.sleep(0.0002)
                else:
                    time.sleep(0.002)
    spotc = LibreSpot()
    tauon.spotc = spotc

    class FFRun:
        def __init__(self):
            self.decoder = None

        def close(self):
            if self.decoder:
                self.decoder.terminate()
            self.decoder = None

        def start(self, uri, start_ms, samplerate):
            self.close()
            path = tauon.get_ffmpeg()
            if not path:
                tauon.test_ffmpeg()
                return 1
            cmd = [path]
            cmd += ["-loglevel", "quiet"]
            if start_ms > 0:
                cmd += ["-ss", f"{start_ms}ms"]
            cmd += ["-i", uri.decode(), "-acodec", "pcm_s16le", "-f", "s16le", "-ac", "2", "-ar", f"{samplerate}", "-"]
            startupinfo = None
            if tauon.msys:
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            try:
                self.decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, startupinfo=startupinfo)
            except:
                print("Error starting ffmpeg")
                return 1
            return 0

        def read(self, buffer: POINTER(c_char), max):
            if self.decoder:
                data = self.decoder.stdout.read(max)
                p = cast(buffer, POINTER(c_char * max))
                p.contents.value = data
                return len(data)
            return 0

    ff_run = FFRun()
    
    def pause_when_device_unavailable():
        pctl.pause_only()
    
    FUNCTYPE = CFUNCTYPE
    if tauon.msys:
        FUNCTYPE = WINFUNCTYPE
    start_callback = FUNCTYPE(c_int, c_char_p, c_int, c_int)(ff_run.start)
    read_callback = FUNCTYPE(c_int, c_void_p, c_int)(ff_run.read)
    close_callback = FUNCTYPE(c_void_p)(ff_run.close)
    device_unavailable_callback = FUNCTYPE(c_void_p)(pause_when_device_unavailable)
    aud.set_callbacks(start_callback, read_callback, close_callback, device_unavailable_callback)

    def calc_rg(track):

        if prefs.replay_gain == 0 and prefs.replay_preamp == 0:
            pctl.active_replaygain = 0
            return 0

        g = 0
        p = 1

        if track is not None:
            tg = track.misc.get("replaygain_track_gain")
            tp = track.misc.get("replaygain_track_peak")
            ag = track.misc.get("replaygain_album_gain")
            ap = track.misc.get("replaygain_album_peak")

            if prefs.replay_gain > 0:
                if prefs.replay_gain == 3 and tg is not None and ag is not None:
                    gens = pctl.gen_codes.get(tauon.pl_to_id(pctl.active_playlist_playing))
                    if pctl.random_mode or (gens and ("st" in gens or "rt" in gens or "r" in gens)):
                        g = tg
                        if tp is not None:
                            p = tp
                    else:
                        g = ag
                        if ap is not None:
                            p = ap
                elif (prefs.replay_gain == 1 and tg is not None) or (prefs.replay_gain == 2 and ag is None and tg is not None):
                    g = tg
                    if tp is not None:
                        p = tp
                elif ag is not None:
                    g = ag
                    if ap is not None:
                        p = ap

        # print("Replay gain")
        # print("GAIN: " + str(g))
        # print("PEAK: " + str(p))
        # print("FINAL: " + str(min(10 ** ((g + prefs.replay_preamp) / 20), 1 / p)))
        if p == 0:
            print("Warning: detected replay gain peak of 0")
            return 1
        pctl.active_replaygain = g
        return min(10 ** ((g + prefs.replay_preamp) / 20), 1 / p)

    audio_cache = tauon.cache_directory + "/network-audio1"
    audio_cache2 = tauon.cache_directory + "/audio-cache"

    class Cachement:
        def __init__(self):
            self.direc = audio_cache2
            if prefs.tmp_cache:
                self.direc = os.path.join(tmp_cache_dir(), "audio-cache")
            if not os.path.exists(self.direc):
                os.makedirs(self.direc)
            self.list = prefs.cache_list
            self.files = os.listdir(self.direc)
            self.get_now = None
            self.running = False
            self.ready = None
            self.error = None

        def get_key(self, track):
            if track.is_network:
                return hashlib.sha256((str(track.index) + track.url_key).encode()).hexdigest()
            else:
                return hashlib.sha256(track.fullpath.encode()).hexdigest()

        def get_file_cached_only(self, track):
            key = self.get_key(track)
            if key in self.files:
                path = os.path.join(self.direc, key)
                if os.path.isfile(path):
                    return path
            return None

        def get_file(self, track):
            # 0: file ready
            # 1: file downloading
            # 2: file not found
            if self.error == track:
                return 2, None

            key = self.get_key(track)
            path = os.path.join(self.direc, key)

            if self.running and self.get_now == track:
                return 1, path

            if key in self.files:
                if os.path.isfile(path):
                    tauon.console.print("got cached file")
                    self.files.remove(key)
                    self.files.append(key)  # bump to top of list
                    self.get_now = None
                    if not self.running:
                        shoot_dl = threading.Thread(target=self.run)
                        shoot_dl.daemon = True
                        shoot_dl.start()
                    return 0, path

            # disable me for debugging
            for codec in (".opus", ".ogg", ".flac", ".mp3"):
                idea = os.path.join(prefs.encoder_output, tauon.encode_folder_name(track), tauon.encode_track_name(track)) + codec
                if os.path.isfile(idea):
                    tauon.console.print("Found transcode")
                    return 0, idea

            self.get_now = track
            if not self.running:
                shoot_dl = threading.Thread(target=self.run)
                shoot_dl.daemon = True
                shoot_dl.start()
            return 1, path

        def run(self):
            self.running = True

            now = self.get_now
            self.get_now = None
            if now is not None:
                error = self.dl_file(now)
                if error:
                    self.error = now
                    self.running = False
                    return

            if self.get_now is None:
                i = 0
                while i < 10:
                    time.sleep(0.1)
                    i += 1
                    if self.get_now is not None:
                        self.running = False
                        return
                tauon.console.print("Precache next track")
                next = pctl.advance(dry=True)
                if next is not None:
                    self.dl_file(pctl.g(next))

            self.trim_cache()
            self.running = False
            return

        def trim_cache(self):
            # Remove untracked items
            for item in self.files:
                t = os.path.join(self.direc, item)
                if os.path.isfile(t) and item not in self.list:
                    os.remove(t)

            # Check total size
            limit = prefs.cache_limit
            if prefs.tmp_cache:
                limit = 10
            while True:
                s = 0
                for item in list(self.list):
                    t = os.path.join(self.direc, item)
                    if not os.path.exists(t):
                        self.list.remove(item)
                        continue
                    s += os.path.getsize(t)
                # Removed oldest items if over limit
                if s > limit * 1000 * 1000 and len(self.list) > 3:
                    t = os.path.join(self.direc, self.list[0])
                    os.remove(t)
                    del self.list[0]
                else:
                    break

        def dl_file(self, track):
            key = self.get_key(track)
            path = os.path.join(self.direc, key)
            if os.path.exists(path):
                if not os.path.isfile(path):
                    return 1
                if key in self.list:
                    return 0
                else:
                    os.remove(path)
            if not track.is_network:
                if not os.path.isfile(track.fullpath):
                    self.error = track
                    self.running = False
                    return 1

                tauon.console.print("start transfer")
                timer = Timer()
                target = open(path, "wb")
                source = open(track.fullpath, "rb")
                while True:
                    try:
                        data = source.read(128000)
                    except:
                        break
                    if len(data) > 0:
                        tauon.console.print(f"Caching file @ {int(len(data) / timer.hit() / 1000)} kbps")
                    else:
                        break
                    target.write(data)
                target.close()
                source.close()
                print("got file")
                self.files.append(key)
                self.list.append(key)
                return 0

            else:
                try:
                    tauon.console.print("Download file")
                    network_url, params = pctl.get_url(track)
                    part = requests.get(network_url, stream=True, params=params, timeout=(3, 10))

                    if part.status_code == 404:
                        gui.show_message("Server: File not found", mode="error")
                        self.error = track
                        return 1
                    elif part.status_code != 200:
                        gui.show_message("Server Error", mode="error")
                        self.error = track
                        return 1

                except:
                    gui.show_message("Could not connect to server", mode="error")
                    self.error = track
                    return 1

                a = 0
                length = 0
                cl = part.headers.get("Content-Length")
                if cl:
                    length = int(cl)
                    gui.buffering_text = "0%"


                timer = Timer()
                try:
                    with open(path, 'wb') as f:
                        for chunk in part.iter_content(chunk_size=1024):
                            if chunk:  # filter out keep-alive new chunks
                                a += 1
                                if a == 1500:  # kilobyes~
                                    self.ready = track
                                if a % 32 == 0:
                                    #time.sleep(0.03)
                                    #tauon.console.print(f"Downloading file @ {round(32 / timer.hit())} kbps")
                                    if length:
                                        gui.update += 1
                                        if True: #a > 2000:
                                            gui.buffering_text = str(round(a * 1000 / length * 100)) + "%"
                                        else:
                                            gui.buffering_text = str(round(a / 2100 * 100)) + "%"

                                if self.get_now is not None and self.get_now != track:
                                    tauon.console.print("ABORT")
                                    return

                                # if self.cancel is True:
                                #     self.part.close()
                                #     self.status = "failed"
                                #     print("Abort download")
                                #     return

                                f.write(chunk)
                    tauon.console.print("DONE")
                    self.files.append(key)
                    self.list.append(key)
                except:
                    print("ERROR")
                    return 1
                return 0


    cachement = Cachement()
    tauon.cachement = cachement

    def set_config(set_device=False):
        aud.config_set_dev_buffer(prefs.device_buffer)
        aud.config_set_fade_duration(prefs.cross_fade_time)
        st = prefs.phazor_device_selected.encode()
        aud.config_set_dev_name(st)
        if set_device:
            aud.pause()
            aud.wait()
            aud.stop_out()
            aud.wait()
            aud.resume()

        if prefs.always_ffmpeg:
            aud.config_set_always_ffmpeg(1)
        if prefs.volume_power < 0 or prefs.volume_power > 10:
            prefs.volume_power = 2
        aud.config_set_volume_power(prefs.volume_power)
        aud.config_set_resample(prefs.avoid_resampling ^ True)

    #aud.config_set_samplerate(prefs.samplerate)
    aud.config_set_resample_quality(prefs.resample)


    set_config()

    def run_vis():
        if gui.turbo:  # and pctl.playing_time > 0.5:
            if gui.vis == 2:
                p_spec = []
                aud.get_spectrum(24, bins1)
                bias = 1
                for b in list(bins1):
                    p_spec.append(int(b * 1.7 * bias))
                    bias += 0.04
                gui.spec = p_spec
                gui.level_update = True
                if pctl.playing_time > 0.5 and (pctl.playing_state == 1 or pctl.playing_state == 3):
                    gui.update_spec = 1
            elif gui.vis == 4:
                p_spec = []
                aud.get_spectrum(45, bins2)
                bias = 1
                for b in list(bins2):
                    p_spec.append(int(b * 2.0 * bias))
                    bias += 0.01
                gui.spec4_array = p_spec
                gui.level_update = True
                if pctl.playing_time > 0.5 and (pctl.playing_state == 1 or pctl.playing_state == 3):
                    gui.update_spec = 1

    stall_timer = Timer()
    wall_timer = Timer()

    def track(end=True):

        run_vis()

        if end and loaded_track and loaded_track.is_network and pctl.playing_time < 7:
            if aud.get_result() == 2:
                print("STALL, RETRY")
                time.sleep(0.5)
                pctl.playerCommandReady = True
                pctl.playerCommand = "open"

        add_time = player_timer.hit()
        if add_time > 2:
            add_time = 2
        if add_time < 0:
            add_time = 0

        pctl.total_playtime += add_time

        ##t = aud.get_position_ms() / 1000
        pctl.playing_time += add_time
        pctl.decode_time = pctl.playing_time

        if pctl.playing_time < 3 and pctl.a_time < 3:
            pctl.a_time = pctl.playing_time
        else:
            pctl.a_time += add_time

        tauon.lfm_scrobbler.update(add_time)

        if len(pctl.track_queue) > 0 and 2 > add_time > 0:
            tauon.star_store.add(pctl.track_queue[pctl.queue_step], add_time)
        if end and pctl.playing_time > 1:
            pctl.test_progress()

    chrome_update = 0
    chrome_cool_timer = Timer()
    chrome_mode = False

    def chrome_start(track, enqueue=False, t=0):
        track = pctl.g(track)
        # if track.is_cue:
        #     print("Error: CUE cast not supported")
        #     return
        if track.is_network:
            if track.file_ext == "SPTY":
                print("Error: Spotify cast not supported")
                return
            network_url, params = pctl.get_url(track)
            if params:
                req = PreparedRequest()
                req.prepare_url(network_url, params)
                network_url = req.url

            tauon.chrome.start(track.index, enqueue=enqueue, url=network_url, t=t)
        else:
            tauon.chrome.start(track.index, enqueue=enqueue, t=t)

    def start_librespot():
        print("SPP: Start librespot command received. Set Phazor input raw mode")
        aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
        spotc.go(force=True)
        if not spotc.running:
            shooter(spotc.worker)

    while True:

        time.sleep(0.016)
        if state == 2:
            time.sleep(0.05)
        if state != 0 or tauon.spot_ctl.playing or tauon.spot_ctl.coasting or tauon.chrome_mode:
            active_timer.set()
        if active_timer.get() > 7:
           aud.stop()
           aud.phazor_shutdown()
           spotc.soft_end()
           break

        # Level meter
        if (state == 1 or state == 3) and gui.vis == 1:
            amp = aud.get_level_peak_l()
            l = amp * 12
            amp = aud.get_level_peak_r()
            r = amp * 12

            tauon.level_train.append((0, l, r))
            gui.level_update = True

        if chrome_mode:
            if pctl.playerCommandReady:
                command = pctl.playerCommand
                # print(command)
                subcommand = pctl.playerSubCommand
                pctl.playerSubCommand = ""
                pctl.playerCommandReady = False

                if command == "endchrome":
                    chrome_mode = False
                    state = 0
                    pctl.playing_time = 0
                    pctl.decode_time = 0
                    pctl.stop()
                    continue
                if command == "open":
                    target_object = pctl.target_object
                    if state == 1:
                        t, pid, s, d = tauon.chrome.update()
                        # print((t, d))
                        # print(d - t)

                        if d and t and 1 < d - t < 5:
                            # print("Enqueue next chromecast")
                            chrome_start(target_object.index, enqueue=True, t=pctl.start_time_target)
                            chrome_cool_timer.set()
                            time.sleep(d - t)
                            if pctl.commit:
                                pctl.advance(quiet=True, end=True)
                                pctl.commit = None
                            continue

                    chrome_start(target_object.index, t=pctl.start_time_target)
                    chrome_cool_timer.set()
                    if pctl.commit:
                        pctl.advance(quiet=True, end=True)
                        pctl.commit = None
                    state = 1
                if command == "pauseon":
                    tauon.chrome.pause()
                    state = 2
                if command == "pauseoff":
                    tauon.chrome.play()
                    state = 1
                if command == "volume":
                    tauon.chrome.volume(round(pctl.player_volume / 100, 3))
                    state = 1
                if command == "seek":
                    tauon.chrome.seek(float(round(pctl.new_time + pctl.start_time_target, 2)))
                    chrome_cool_timer.set()
                    pctl.playing_time = pctl.new_time
                    pctl.decode_time = pctl.playing_time
                if command == "stop":
                    state = 0
                    tauon.chrome.stop()

            if state == 1:

                if chrome_update > 0.8 and chrome_cool_timer.get() > 2.5:
                    t, pid, s, d = tauon.chrome.update()
                    pctl.playing_time = t - pctl.start_time_target
                    pctl.decode_time = t - pctl.start_time_target
                    player_timer.hit()
                    chrome_update = 0

                add_time = player_timer.hit()
                if add_time > 2:
                    add_time = 2
                chrome_update += add_time
                pctl.a_time += add_time
                tauon.lfm_scrobbler.update(add_time)
                pctl.total_playtime += add_time
                if len(pctl.track_queue) > 0 and 2 > add_time > 0:
                    tauon.star_store.add(pctl.track_queue[pctl.queue_step], add_time)

                pctl.test_progress()

            time.sleep(0.1)
            continue

        # Command processing
        if pctl.playerCommandReady:

            command = pctl.playerCommand
            subcommand = pctl.playerSubCommand
            pctl.playerSubCommand = ""
            pctl.playerCommandReady = False
            #print(command)

            if command == "spotcon":

                #aud.stop()
                print("SPP: Start librespot command received. Set Phazor input raw mode")
                start_librespot()
                state = 4
                #time.sleep(20)
                #tauon.spot_ctl.preparing_spotify = False

            if command == "startchrome":
                aud.stop()
                if state == 1:
                    chrome_start(loaded_track.index, t=pctl.playing_time)
                chrome_mode = True

            if command == "reload":
                set_config()
            if command == "set-device":
                set_config(set_device=True)

            if command == "url":
                pctl.download_time = 0
                w = 0
                if not tauon.radiobox.run_proxy:
                    aud.start(pctl.url.encode(), 0, 0, ctypes.c_float(calc_rg(None)))
                    state = 3
                    player_timer.hit()
                else:
                    while len(tauon.stream_proxy.chunks) < 200:
                        time.sleep(0.1)
                        w += 1
                        if w > 100:
                            print("Taking too long!")
                            tauon.stream_proxy.stop()
                            pctl.playerCommand = 'stop'
                            pctl.playerCommandReady = True
                            break
                    else:
                        aud.config_set_feed_samplerate(prefs.samplerate)
                        aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
                        state = 3
                        player_timer.hit()

            if command == "open":
                if state == 2:
                    aud.set_volume(int(pctl.player_volume))


                stall_timer.set()
                pctl.download_time = 0
                target_object = pctl.target_object
                target_path = target_object.fullpath
                subtrack = target_object.subtrack
                aud.set_subtrack(subtrack)

                tauon.console.print(f"open - requested start was {int(pctl.start_time_target + pctl.jump_time)} ({pctl.start_time_target})")
                try:
                    tauon.console.print(target_path.split(".")[1])
                except:
                    pass

                if (tauon.spot_ctl.playing or tauon.spot_ctl.coasting) and not target_object.file_ext == "SPTY":
                    tauon.spot_ctl.control("stop")

                if target_object.is_network:

                    if target_object.file_ext == "SPTY":
                        tauon.level_train.clear()
                        if target_object.found is False:
                            pctl.playing_state = 0
                            pctl.jump_time = 0
                            pctl.advance(inplace=True, play=True)
                            continue
                        if state > 0 and not state == 4:
                            aud.stop()
                        if not state == 4:
                            state = 0

                        try:
                            f = False
                            if spotc.running and state != 4:
                                aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
                                state = 4
                            if prefs.launch_spotify_local and not spotc.running:
                                aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
                                state = 4
                                if not tauon.librespot_p:
                                    tauon.spot_ctl.preparing_spotify = True
                                    f = True
                                spotc.go()
                                if not spotc.running:
                                    shooter(spotc.worker)

                            tauon.spot_ctl.play_target(target_object.url_key, force_new_device=f, start_callback=start_librespot)
                            #tauon.spot_ctl.preparing_spotify = False
                            if pctl.playerCommand == "spotcon":  # such spaghetti code
                                state = 4
                                pctl.playerCommand = ""
                                pctl.playerCommandReady = False

                        except:
                            #tauon.spot_ctl.preparing_spotify = False
                            print("Failed to start Spotify track")
                            pctl.playerCommand = "stop"
                            pctl.playerCommandReady = True
                        continue

                    spotc.running = False
                    #tauon.spot_ctl.preparing_spotify = False
                    pctl.spot_playing = False
                    timer = Timer()
                    timer.set()
                    while True:
                        status, path = cachement.get_file(target_object)

                        if status == 0 or status == 2:
                            break
                        if timer.get() > 0.25 and gui.buffering is False:
                            gui.buffering_text = ""
                            gui.buffering = True
                            gui.update += 1
                            tauon.wake()
                        if cachement.ready == target_object and pctl.start_time_target + pctl.jump_time == 0:
                            break
                        time.sleep(0.05)
                        #print(status)

                    gui.buffering = False
                    gui.update += 1
                    tauon.wake()

                    if status == 2:
                        tauon.console.print("Could not locate resource")
                        target_object.found = False
                        pctl.playing_state = 0
                        pctl.jump_time = 0
                        #pctl.advance(inplace=True, play=True)
                        continue
                    target_path = path

                elif prefs.precache:
                    timer = Timer()
                    timer.set()
                    while True:
                        status, path = cachement.get_file(target_object)
                        if status == 0 or status == 2:
                            break
                        if timer.get() > 0.25 and gui.buffering is False:
                            gui.buffering_text = ""
                            gui.buffering = True
                            gui.update += 1
                            tauon.wake()

                        time.sleep(0.05)

                    gui.buffering = False
                    gui.update += 1
                    tauon.wake()

                    if status == 2:
                        target_object.found = False
                        pctl.playing_state = 0
                        pctl.jump_time = 0
                        pctl.advance(inplace=True, play=True)
                        continue
                    target_path = path

                if not os.path.isfile(target_path):
                    target_object.found = False
                    if not target_object.is_network:
                        pctl.playing_state = 0
                        pctl.jump_time = 0
                        pctl.advance(inplace=True, play=True)
                    continue
                elif not target_object.found:
                    pctl.reset_missing_flags()

                spotc.running = False
                length = 0
                remain = 0
                position = 0

                if target_path and target_object and target_object.length == 0 and not target_object.is_cue:
                    print("Track has duration of 0, scanning file")
                    temp = tauon.TrackClass()
                    temp.fullpath = target_path
                    tauon.tag_scan(temp)
                    target_object.length = temp.length
                    if pctl.playing_object() is target_object:
                        pctl.playing_length = target_object.length
                    del temp

                if state == 1 and not pctl.start_time_target and not pctl.jump_time and \
                        loaded_track:

                    length = aud.get_length_ms() / 1000
                    position = aud.get_position_ms() / 1000
                    remain = length - position

                    tauon.console.print(loaded_track.title + " -> " + target_object.title)
                    tauon.console.print(" --- length: " + str(length))
                    tauon.console.print(" --- position: " + str(position))
                    tauon.console.print(" --- We are %s from end" % str(remain))

                    if loaded_track.is_network or length == 0:
                        tauon.console.print("Phazor did not respond with a duration")
                        length = loaded_track.length
                        remain = length - position

                fade = 0
                error = False
                if state == 1 and length and position and not pctl.start_time_target and not pctl.jump_time and \
                        loaded_track and 0 < remain < 5.5 and not loaded_track.is_cue and subcommand != "now":

                    tauon.console.print("Transition gapless")

                    r_timer = Timer()
                    r_timer.set()

                    if loaded_track and loaded_track.file_ext.lower() in tauon.gme_formats:
                        # GME formats dont have a physical end so we don't do gapless
                        while r_timer.get() <= remain - prefs.device_buffer / 1000:
                            if pctl.commit:
                                track(end=False)
                            time.sleep(0.016)
                        aud.stop()

                    aud.next(target_path.encode(), int(pctl.start_time_target + pctl.jump_time) * 1000, ctypes.c_float(calc_rg(target_object)))

                    cont = False
                    while r_timer.get() <= remain - prefs.device_buffer / 1000:
                        if pctl.commit:
                            track(end=False)
                        time.sleep(0.016)
                        if pctl.playerCommandReady and pctl.playerCommand in ("open", "stop"):
                            print("JANK")
                            pctl.commit = None
                            break
                        if pctl.playerCommandReady and pctl.playerCommand == "seek":
                            print("Seek revert gapless")
                            pctl.commit = None
                            pctl.jump(loaded_track.index, jump=True)
                            pctl.jump_time = pctl.new_time
                            pctl.playing_time = pctl.new_time
                            pctl.decode_time = pctl.new_time
                            cont = True
                            break

                    if cont:
                        continue
                    if pctl.commit is not None:
                        pctl.playing_time = 0
                        pctl.decode_time = 0
                        match = pctl.commit
                        if match == loaded_track.index and subcommand == "repeat":
                            pctl.update_change()
                        else:
                            pctl.advance(quiet=True, end=True)
                        pt = pctl.playing_object()
                        if pt and pt.index != match:
                            print("MISSFIRE")
                            pctl.play_target()
                            continue
                        elif pctl.playerCommandReady and pctl.playerCommand == "open":
                            pctl.playerCommandReady = False
                            pctl.playerCommand = ""

                    loaded_track = target_object
                    pctl.playing_time = pctl.jump_time

                else:
                    if pctl.commit:
                        pctl.advance(quiet=True, end=True)
                        pctl.commit = None
                        continue

                    if state == 1 and prefs.use_jump_crossfade:
                        fade = 1

                    tauon.console.print("Transition jump")
                    aud.start(target_path.encode(errors='surrogateescape'), int(pctl.start_time_target + pctl.jump_time) * 1000, fade, ctypes.c_float(calc_rg(target_object)))
                    loaded_track = target_object
                    pctl.playing_time = pctl.jump_time
                    if pctl.jump_time:
                        while aud.get_result() == 0:
                            time.sleep(0.016)
                            run_vis()
                        aud.set_position_ms(int(pctl.jump_time * 1000))

                    # Restart track is failed to load (for some network tracks) (broken with gapless loading)
                    while True:
                        r = aud.get_result()
                        if r == 1:
                            break
                        if r == 2:
                            if loaded_track.is_network:
                                gui.buffering = True
                                gui.buffering_text = ""

                                # while dm.request(loaded_track, whole=True) == "wait":
                                #     time.sleep(0.05)
                                #     if pctl.playerCommandReady:
                                #         break
                                print("Retry start file")
                                aud.start(target_path.encode(), int(pctl.start_time_target + pctl.jump_time) * 1000,
                                          fade, ctypes.c_float(calc_rg(target_object)))
                                gui.buffering = False
                                player_timer.set()
                                break
                            else:
                                aud.stop()
                                if not gui.message_box:
                                    gui.show_message("Error loading track", mode="warning")
                                error = True
                                break
                        time.sleep(0.016)
                        run_vis()

                    state = 1
                    if error:
                        state = 0

                player_timer.set()
                pctl.jump_time = 0
                if loaded_track.length == 0 or loaded_track.file_ext.lower() in tauon.mod_formats:
                    i = 0
                    t = 0
                    while t == 0:
                        time.sleep(0.3)
                        t = aud.get_length_ms() / 1000
                        i += 1
                        if i > 9:
                            break
                    loaded_track.length = t
                    if loaded_track.length != 0:
                        pctl.playing_length = loaded_track.length
                        gui.pl_update += 1

                pctl.commit = None
                stall_timer.set()
                wall_timer.force_set(3)

            if command == "seek":
                if tauon.spot_ctl.coasting or tauon.spot_ctl.playing:
                    tauon.spot_ctl.control("seek", int(pctl.new_time * 1000))
                    pctl.playing_time = pctl.new_time
                elif state > 0:

                    if loaded_track.is_network:  # and loaded_track.fullpath.endswith(".ogg"):

                        timer = Timer()
                        timer.set()
                        while True:
                            status, path = cachement.get_file(loaded_track)
                            if status == 0 or status == 2:
                                break
                            if timer.get() > 0.25 and gui.buffering is False:
                                gui.buffering_text = ""
                                gui.buffering = True
                                gui.update += 1
                                tauon.wake()

                            time.sleep(0.05)

                        gui.buffering = False
                        gui.update += 1
                        tauon.wake()

                        if status == 2:
                            loaded_track.found = False
                            pctl.playing_state = 0
                            pctl.jump_time = 0
                            pctl.stop()
                            continue

                        # The vorbis decoder doesn't like appended files
                        aud.start(path.encode(), int(pctl.new_time + pctl.start_time_target) * 1000, 0, ctypes.c_float(calc_rg(loaded_track)))
                        while aud.get_result() == 0:
                            time.sleep(0.01)
                    else:
                        aud.seek(int((pctl.new_time + pctl.start_time_target) * 1000), prefs.pa_fast_seek)

                    pctl.playing_time = pctl.new_time

                pctl.decode_time = pctl.playing_time
                wall_timer.set()

            if command == "volume":
                if (tauon.spot_ctl.coasting or tauon.spot_ctl.playing) and not spotc.running:
                    tauon.spot_ctl.control("volume", int(pctl.player_volume))
                else:
                    aud.ramp_volume(int(pctl.player_volume), 750)

            if command == "runstop":
                length = aud.get_length_ms() / 1000
                position = aud.get_position_ms() / 1000
                remain = length - position
                # print("length: " + str(length))
                # print("position: " + str(position))
                # print("We are %s from end" % str(remain))
                time.sleep(3)
                command = "stop"

            if command == "stop":

                if prefs.use_pause_fade and state != 3:
                    if pctl.player_volume > 5:
                        speed = fade_time / (int(pctl.player_volume) / 100)
                    else:
                        speed = fade_time / (5 / 100)

                    aud.ramp_volume(0, int(speed))
                    time.sleep((fade_time + 100) / 1000)

                state = 0
                pctl.playing_time = 0
                aud.stop()
                spotc.running = False
                time.sleep(0.1)
                aud.set_volume(int(pctl.player_volume))

                if subcommand == "return":
                    pctl.playerSubCommand = "stopped"
                    #pctl.playerCommandReady = True

            if command == "pauseon":
                if prefs.use_pause_fade:
                    if pctl.player_volume > 5:
                        speed = fade_time / (int(pctl.player_volume) / 100)
                    else:
                        speed = fade_time / (5 / 100)

                    aud.ramp_volume(0, int(speed))
                    time.sleep((fade_time + 100) / 1000)
                aud.pause()
                state = 2

            if command == "pauseoff":
                if prefs.use_pause_fade:
                    if pctl.player_volume > 5:
                        speed = fade_time / (int(pctl.player_volume) / 100)
                    else:
                        speed = fade_time / (5 / 100)

                    aud.ramp_volume(int(pctl.player_volume), int(speed))
                aud.resume()
                player_timer.set()
                stall_timer.set()
                state = 1

            if command == "unload":

                if prefs.use_pause_fade:
                    if pctl.player_volume > 5:
                        speed = fade_time / (int(pctl.player_volume) / 100)
                    else:
                        speed = fade_time / (5 / 100)

                    aud.ramp_volume(0, int(speed))
                    time.sleep((fade_time + 100) / 1000)

                spotc.end()
                aud.stop()
                aud.phazor_shutdown()

                if os.path.exists(audio_cache):
                    shutil.rmtree(audio_cache)

                pctl.playerCommand = "done"
                pctl.playerCommandReady = True
                break
        else:
            pctl.spot_test_progress()

            if state == 4:
                run_vis()

            if state == 3:

                pctl.radio_progress()
                run_vis()

                add_time = player_timer.hit()
                pctl.playing_time += add_time
                pctl.decode_time = pctl.playing_time

                buffering = aud.is_buffering()
                if gui.buffering != buffering:
                    gui.buffering = buffering
                    gui.update += 1

            if state == 1:
                track()


